from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format
from keras_cv_attention_models.attention_layers import (
    add_with_layer_scale_and_drop_block,
    batchnorm_with_activation,
    conv2d_no_bias,
    depthwise_conv2d_no_bias,
    layer_norm,
    mlp_block,
    mlp_mixer_block,
    output_block,
    PositionalEmbedding,
    qkv_to_multi_head_channels_last_format,
    scaled_dot_product_attention,
    window_attention,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

LAYER_NORM_EPSILON = 1e-6

PRETRAINED_DICT = {
    "gpvit_l1": {"imagenet": {224: "4c2b124acdb20cc9ef33a32a85a2cd4e"}},
    "gpvit_l2": {"imagenet": {224: "ce5e72b80cfcb9a9567e8e52d42b4e15"}},
    "gpvit_l3": {"imagenet": {224: "a378001b60878ea8851ebe78a28bcfbe"}},
    "gpvit_l4": {"imagenet": {224: "0cde8fcea39794ea0ce1ffdf7c49eef0"}},
}


@backend.register_keras_serializable(package="kecam")
class PureWeigths(layers.Layer):
    """Just return a weights with specific shape"""

    def __init__(self, shape, **kwargs):
        super().__init__(**kwargs)
        self.shape = shape

    def build(self, input_shape):
        self.gain = self.add_weight(name="gain", shape=self.shape, dtype="float32", trainable=True)
        super().build(input_shape)

    def call(self, inputs, **kwargs):
        return self.gain

    def get_config(self):
        config = super().get_config()
        config.update({"shape": self.shape})
        return config

    def compute_output_shape(self, input_shape):
        return self.shape


def lepe_attention(inputs, num_heads=4, dropout=0, name=""):
    query, key, value = functional.split(inputs, 3, axis=-1)
    _, hh, ww, input_channel = query.shape
    blocks = hh * ww
    key_dim = input_channel // num_heads
    # print(f"{input_channel = }, {num_heads = }, {key_dim = }")

    lepe = value if image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(value)  # channels_last -> channels_first
    lepe = depthwise_conv2d_no_bias(lepe, kernel_size=3, use_bias=True, padding="same", name=name + "lepe_")
    lepe = lepe if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(lepe)  # channels_first -> channels_last

    query, key, value = qkv_to_multi_head_channels_last_format(query, key, value, num_heads=num_heads, data_format="channels_last")
    # print(f"{query.shape = }, {key.shape = }, {value.shape = }, {lepe.shape = }")

    output_shape = (hh, ww, input_channel)
    attn_out = scaled_dot_product_attention(query, key, value, output_shape=output_shape, out_weight=False, dropout=dropout, name=name)
    return layers.Add()([attn_out, lepe])


def window_lepe_attention(inputs, num_heads=4, window_size=2, key_dim=0, qkv_bias=True, dropout=0, name=""):
    _, hh, ww, input_channel = inputs.shape
    key_dim = key_dim if key_dim > 0 else input_channel // num_heads
    # out_shape = input_channel if out_shape is None or not out_weight else out_shape
    emb_dim = num_heads * key_dim

    # qkv = self.qkv(img).reshape(B, -1, 3, C).permute(2, 0, 1, 3).contiguous()
    # split 3 -> split 2
    qkv = layers.Dense(emb_dim * 3, use_bias=qkv_bias, name=name and name + "qkv")(inputs)
    qkv = functional.reshape(qkv, [-1, hh * ww * 3, emb_dim])
    qkv_left, qkv_right = functional.split(qkv, 2, axis=-1)
    qkv_left = functional.reshape(qkv_left, [-1, hh, ww, 3 * qkv_left.shape[-1]])
    qkv_right = functional.reshape(qkv_right, [-1, hh, ww, 3 * qkv_right.shape[-1]])

    attn_out_left = window_attention(qkv_left, window_size=(hh, window_size), num_heads=num_heads // 2, attention_block=lepe_attention, name=name + "left_")
    attn_out_right = window_attention(qkv_right, window_size=(window_size, ww), num_heads=num_heads // 2, attention_block=lepe_attention, name=name + "right_")
    attn_out = functional.concat([attn_out_left, attn_out_right], axis=-1)
    # print(f"{inputs.shape = }, {attn_out_left.shape = }, {attn_out_right.shape = }, {attn_out.shape = }")
    attn_out = layers.Dense(attn_out.shape[-1], use_bias=True, name=name and name + "attn_out")(attn_out)
    return attn_out


def window_lepe_attention_mlp_block(inputs, num_heads=4, window_size=2, mlp_ratio=4, layer_scale=0.1, drop_rate=0, activation="gelu", name=""):
    input_channel = inputs.shape[-1]  # Channels_last only

    """ attention """
    nn = layer_norm(inputs, epsilon=LAYER_NORM_EPSILON, axis=-1, name=name + "attn_")
    nn = window_lepe_attention(nn, num_heads=num_heads, window_size=window_size, name=name + "attn_")
    attn_out = add_with_layer_scale_and_drop_block(inputs, nn, layer_scale=layer_scale, drop_rate=drop_rate, axis=-1, name=name + "attn_")

    """ MLP """
    nn = layer_norm(attn_out, epsilon=LAYER_NORM_EPSILON, axis=-1, name=name + "mlp_")
    nn = mlp_block(nn, input_channel * mlp_ratio, activation=activation, name=name + "mlp_")
    nn = add_with_layer_scale_and_drop_block(attn_out, nn, layer_scale=layer_scale, drop_rate=drop_rate, axis=-1, name=name + "mlp_")

    """ DepthWise """
    nn = nn if image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(nn)  # channels_last -> channels_first
    nn = depthwise_conv2d_no_bias(nn, kernel_size=3, padding="same", name=name + "output_")
    nn = nn if image_data_format() == "channels_last" else layers.Permute([2, 3, 1], name=name + "output_perm")(nn)  # channels_first -> channels_last
    return nn


def light_group_attention(inputs, num_heads=4, key_dim=0, num_group_token=0, use_key_value_norm=True, qkv_bias=True, dropout=0, name=""):
    input_channel = inputs.shape[-1]
    key_dim = key_dim if key_dim > 0 else input_channel // num_heads

    query = PureWeigths(shape=[1, num_group_token, input_channel], name=name + "query")(inputs)
    query = layer_norm(query, epsilon=LAYER_NORM_EPSILON, axis=-1, name=name + "query_")
    key_value = layer_norm(inputs, epsilon=LAYER_NORM_EPSILON, axis=-1, name=name + "key_value_") if use_key_value_norm else inputs
    key = layers.Dense(key_value.shape[-1], use_bias=qkv_bias, name=name and name + "key")(key_value)
    # print(f"{inputs.shape = }, {query.shape = }, {key.shape = }, {key_value.shape = }, {num_heads = }")
    query, key, value = qkv_to_multi_head_channels_last_format(query, key, key_value, num_heads=num_heads, data_format="channels_last")

    output_shape = [num_group_token, input_channel]
    return scaled_dot_product_attention(query, key, value, output_shape=output_shape, out_weight=False, dropout=dropout, name=name)


def full_ungroup_attention(inputs, group_token, num_heads=4, key_dim=0, qkv_bias=True, dropout=0, name=""):
    # inputs: [batch, height, width, channel], group_token: [batch, num_group_token, channel]
    input_channel = inputs.shape[-1]
    key_dim = key_dim if key_dim > 0 else input_channel // num_heads

    query = layer_norm(inputs, epsilon=LAYER_NORM_EPSILON, axis=-1, name=name + "query_")
    key_value = layer_norm(group_token, epsilon=LAYER_NORM_EPSILON, axis=-1, name=name + "key_value_")

    query = layers.Dense(query.shape[-1], use_bias=qkv_bias, name=name and name + "query")(query)
    key = layers.Dense(key_value.shape[-1], use_bias=qkv_bias, name=name and name + "key")(key_value)
    value = layers.Dense(key_value.shape[-1], use_bias=qkv_bias, name=name and name + "value")(key_value)
    query, key, value = qkv_to_multi_head_channels_last_format(query, key, value, num_heads=num_heads, data_format="channels_last")

    attn = scaled_dot_product_attention(query, key, value, output_shape=inputs.shape, out_weight=False, dropout=dropout, name=name)
    attn = layers.Dense(input_channel, use_bias=True, name=name and name + "out")(attn)

    attn_out = functional.concat([inputs, attn], axis=-1)
    attn_out = layers.Dense(input_channel, use_bias=True, name=name and name + "attn_out")(attn_out)
    return attn_out


def group_attention(inputs, num_heads=4, num_group_token=0, mlp_ratio=4, layer_scale=0.1, drop_rate=0, activation="gelu", name=""):
    input_channel = inputs.shape[-1]  # Channels_last only
    group_token = light_group_attention(inputs, num_heads=num_heads, num_group_token=num_group_token, name=name + "light_attn_")

    tokens_mlp_dim, channels_mlp_dim = input_channel // 2, input_channel * 4  # all using embed_dims
    group_token = mlp_mixer_block(
        group_token, tokens_mlp_dim, channels_mlp_dim, drop_rate=drop_rate, activation=activation, data_format="channels_last", name=name
    )

    attn_out = full_ungroup_attention(inputs, group_token, num_heads=num_heads, name=name + "full_attn_")
    attn_out = layers.Reshape(inputs.shape[1:])(attn_out)

    """ MLP """
    nn = layer_norm(attn_out, axis=-1, epsilon=LAYER_NORM_EPSILON, name=name + "mlp_")
    nn = mlp_block(nn, input_channel * mlp_ratio, activation=activation, name=name + "mlp_")
    nn = add_with_layer_scale_and_drop_block(attn_out, nn, layer_scale=layer_scale, drop_rate=drop_rate, axis=-1, name=name + "mlp_")

    """ DepthWise """
    nn = nn if image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(nn)  # channels_last -> channels_first
    nn = depthwise_conv2d_no_bias(nn, kernel_size=3, padding="same", name=name)
    nn = batchnorm_with_activation(nn, activation="relu", name=name + "output_")
    nn = nn if image_data_format() == "channels_last" else layers.Permute([2, 3, 1], name=name + "output_perm")(nn)  # channels_first -> channels_last
    return nn


def GPViT(
    num_layers=12,
    embed_dims=216,
    stem_depth=1,
    num_window_heads=12,
    num_group_heads=6,
    mlp_ratios=4,
    window_size=2,
    group_attention_layer_ids=[1, 4, 7, 10],
    group_attention_layer_group_tokens=[64, 32, 32, 16],
    use_neck_attention_output=True,
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="gelu",
    drop_connect_rate=0,
    dropout=0,
    layer_scale=0,
    classifier_activation="softmax",
    pretrained=None,
    model_name="gp_vit",
    kwargs=None,
):
    # Regard input_shape as force using original shape if first element is None or -1,
    # else assume channel dimention is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)

    """ Stem """
    nn = conv2d_no_bias(inputs, 64, kernel_size=7, strides=2, padding="same", name="stem_")
    nn = batchnorm_with_activation(nn, activation="relu", name="stem_")
    for id in range(stem_depth - 1):
        nn = conv2d_no_bias(nn, 64, kernel_size=3, strides=1, padding="same", name="stem_{}_".format(id + 1))
        nn = batchnorm_with_activation(nn, activation="relu", name="stem_{}_".format(id + 1))
    ## nn = AdaptivePadding(nn) with padding='corner' --> use_torch_padding=False
    nn = conv2d_no_bias(nn, embed_dims, kernel_size=4, strides=4, padding="same", use_torch_padding=False, use_bias=True, name="stem_patch_")
    nn = nn if image_data_format() == "channels_last" else layers.Permute([2, 3, 1], name="stem_perm")(nn)  # channels_first -> channels_last
    nn = PositionalEmbedding(name="positional_embedding")(nn)  # Needs to be channels_last for compitable with TF

    """ stacks """
    group_attention_layer_group_tokens = group_attention_layer_group_tokens.copy()
    global_block_id = 0
    for block_id in range(num_layers):
        block_name = "block{}_".format(block_id + 1)
        block_drop_rate = drop_connect_rate * global_block_id / num_layers
        if block_id in group_attention_layer_ids:
            num_group_token = group_attention_layer_group_tokens.pop(0)
            nn = group_attention(nn, num_group_heads, num_group_token, mlp_ratios, layer_scale, block_drop_rate, activation=activation, name=block_name)
        else:
            nn = window_lepe_attention_mlp_block(nn, num_window_heads, window_size, mlp_ratios, layer_scale, block_drop_rate, activation, name=block_name)
        global_block_id += 1

    nn = layer_norm(nn, epsilon=LAYER_NORM_EPSILON, axis=-1, name="pre_out_")
    if use_neck_attention_output:
        nn = light_group_attention(nn, num_heads=6, num_group_token=1, use_key_value_norm=False, qkv_bias=False, name="neck_")
        nn = layers.Flatten()(nn)
    else:
        nn = nn if image_data_format() == "channels_last" else layers.Permute([3, 1, 2], name="out_perm")(nn)  # channels_last -> channels_first

    nn = output_block(nn, num_classes=num_classes, drop_rate=dropout, classifier_activation=classifier_activation)
    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, PRETRAINED_DICT, "gpvit", pretrained, PositionalEmbedding)
    return model


def GPViT_L1(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return GPViT(**locals(), model_name="gpvit_l1", **kwargs)


def GPViT_L2(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dims = 348
    stem_depth = 2
    return GPViT(**locals(), model_name="gpvit_l2", **kwargs)


def GPViT_L3(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dims = 432
    stem_depth = 2
    use_neck_attention_output = False
    return GPViT(**locals(), model_name="gpvit_l3", **kwargs)


def GPViT_L4(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dims = 624
    stem_depth = 3
    use_neck_attention_output = False
    return GPViT(**locals(), model_name="gpvit_l4", **kwargs)



from keras_cv_attention_models.coat.coat import CoaT, CoaTLiteTiny, CoaTLiteMini, CoaTLiteSmall, CoaTTiny, CoaTMini, ConvPositionalEncoding, ConvRelativePositionalEncoding, ClassToken

__head_doc__ = """
Keras implementation of [Github mlpc-ucsd/CoaT](https://github.com/mlpc-ucsd/CoaT).
Paper [PDF 2104.06399 CoaT: Co-Scale Conv-Attentional Image Transformers](http://arxiv.org/abs/2104.06399).
"""

__tail_doc__ = """  head_splits: split head list for `ConvRelativePositionalEncoding`.
      Should be sum eqauls `num_heads`. Default `[2, 3, 3]` indicates split `8` head into 3 groups.
  head_kernel_size: kernel_size for each split head in `ConvRelativePositionalEncoding`. Defualt `[3, 5, 7]`.
  use_shared_cpe: set `False` to disable using shared `ConvPositionalEncoding` blocks.
  use_shared_crpe: set `False` to disable using shared `ConvRelativePositionalEncoding` blocks.
  out_features: None or a list of number in `[0, 1, 2, 3]` for output of relative `serial block` output.
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `gelu`.
  drop_connect_rate: Not using.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
  pretrained: one of `None` (random initialization) or 'imagenet' (pre-training on ImageNet).
      Will try to download and load pre-trained model weights if not None.
  **kwargs: other parameters if available.

Returns:
    A `keras.Model` instance.
"""

CoaT.__doc__ = __head_doc__ + """
Args:
  serial_depths: number of serial blocks.
  embed_dims: output channel for each serial stack.
  mlp_ratios: dimension expansion ration for `mlp_block` in each `serial` and `parallel` stack.
  parallel_depth: number of parallel blocks. `0` for tiny models, `6` for none tiny models.
  patch_size: patch size extracted from input for all `patch_embed` blocks.
  num_heads: heads number for transformer block.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model         | Params | FLOPs | Input | Top1 Acc |
  | ------------- | ------ | ----- | ----- | -------- |
  | CoaTLiteTiny  | 5.7M   | 1.60G | 224   | 77.5     |
  | CoaTLiteMini  | 11M    | 2.00G | 224   | 79.1     |
  | CoaTLiteSmall | 20M    | 3.97G | 224   | 81.9     |
  | CoaTTiny      | 5.5M   | 4.33G | 224   | 78.3     |
  | CoaTMini      | 10M    | 6.78G | 224   | 81.0     |
"""

CoaTLiteTiny.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__
CoaTLiteMini.__doc__ = CoaTLiteTiny.__doc__
CoaTLiteSmall.__doc__ = CoaTLiteTiny.__doc__
CoaTTiny.__doc__ = CoaTLiteTiny.__doc__
CoaTMini.__doc__ = CoaTLiteTiny.__doc__

ConvPositionalEncoding.__doc__ = __head_doc__ + """
Convolutional Position Encoding. Note: This module is similar to the conditional position encoding in CPVT.
Applies a `DepthwiseConv2D` layer with input, then adds with input.
Not a layer, just wappered a class for reusable.

input: `[batch, class_token + height * width, channel]`.
output: `[batch, class_token + height * width, channel]`.

Args:
  kernel_size: `DepthwiseConv2D` kernel size.

Examples:
>>> from keras_cv_attention_models import attention_layers
>>> aa = attention_layers.ConvPositionalEncoding()
>>> print(f"{aa(tf.ones([1, 1 + 14 * 14, 256])).shape = }")
# aa(tf.ones([1, 1 + 14 * 14, 256])).shape = TensorShape([1, 197, 256])
"""

ConvRelativePositionalEncoding.__doc__ = __head_doc__ + """
Convolutional with Relative Position Encoding.
Applies multi `DepthwiseConv2D` layers with split input, then adds with input.
Not a layer, just wappered a class for reusable.

input:
    query: `[batch, num_heads, class_token + height * width, channel // num_heads]`.
    value: `[batch, num_heads, class_token + height * width, channel // num_heads]`.
output: `[batch, num_heads, 1 * zero + height * width, channel // num_heads]`.

Args:
  head_splits: split head list. Should be sum eqauls `num_heads`. Default `[2, 3, 3]` indicates split `8` head into 3 groups.
  head_kernel_size: kernel_size for each split head. Defualt `[3, 5, 7]`.

Examples:
>>> from keras_cv_attention_models import attention_layers
>>> aa = attention_layers.ConvRelativePositionalEncoding()
>>> print(f"{aa(tf.ones([1, 8, 1 + 14 * 14, 6]), tf.ones([1, 8, 1 + 14 * 14, 6])).shape = }")
# aa(tf.ones([1, 8, 1 + 14 * 14, 6]), tf.ones([1, 8, 1 + 14 * 14, 6])).shape = TensorShape([1, 8, 197, 6])
"""



import os
import numpy as np
from keras_cv_attention_models.backend import layers, models, functional, image_data_format, get_file


def reload_model_weights(
    model, pretrained_dict, sub_release, pretrained="imagenet", mismatch_class=None, force_reload_mismatch=False, request_resolution=-1, method=None
):
    if not isinstance(pretrained, str):
        return
    if pretrained.endswith(".h5"):
        print(">>>> Load pretrained from:", pretrained)
        # model.load_weights(pretrained, by_name=True, skip_mismatch=True)
        load_weights_with_mismatch(model, pretrained, mismatch_class, request_resolution=request_resolution, method=method)
        return pretrained

    file_hash = pretrained_dict.get(model.name, {}).get(pretrained, None)
    if file_hash is None:
        print(">>>> No pretrained available, model will be randomly initialized")
        return None

    if isinstance(file_hash, dict):
        # file_hash is a dict like {224: "aa", 384: "bb", 480: "cc"}
        if request_resolution == -1:
            input_height = model.input_shape[2]  # Either channels_last or channels_first, 2 aixs has to be a shape one
            if input_height is None:  # input_shape is (None, None, 3)
                request_resolution = max(file_hash.keys())
            else:
                request_resolution = min(file_hash.keys(), key=lambda ii: abs(ii - input_height))
        pretrained = "{}_".format(request_resolution) + pretrained
        file_hash = file_hash[request_resolution]
        # print(f"{request_resolution = }, {pretrained = }, {file_hash = }")
    elif request_resolution == -1:
        request_resolution = 224  # Default is 224

    pre_url = "https://github.com/leondgarse/keras_cv_attention_models/releases/download/{}/{}_{}.h5"
    url = pre_url.format(sub_release, model.name, pretrained)
    file_name = os.path.basename(url)
    try:
        pretrained_model = get_file(file_name, url, cache_subdir="models", file_hash=file_hash)
    except:
        print("[Error] will not load weights, url not found or download failed:", url)
        return None
    else:
        print(">>>> Load pretrained from:", pretrained_model)
        # model.load_weights(pretrained_model, by_name=True, skip_mismatch=True)
        load_weights_with_mismatch(model, pretrained_model, mismatch_class, force_reload_mismatch, request_resolution, method)
        return pretrained_model


def load_weights_with_mismatch(model, weight_file, mismatch_class=None, force_reload_mismatch=False, request_resolution=-1, method=None):
    model.load_weights(weight_file, by_name=True, skip_mismatch=True)
    if image_data_format() == "channels_first":
        input_height, input_width = model.input_shape[2], model.input_shape[3]
    else:
        input_height, input_width = model.input_shape[1], model.input_shape[2]

    # if mismatch_class is not None:
    if mismatch_class is not None and (force_reload_mismatch or request_resolution != input_height or request_resolution != input_width):
        try:
            import h5py

            print(">>>> Reload mismatched weights: {} -> {}".format(request_resolution, (input_height, input_width)))
            with h5py.File(weight_file, mode="r") as h5_file:
                weights = h5_file["model_weights"] if "model_weights" in h5_file else h5_file  # full model or weights only

                if isinstance(mismatch_class, (list, tuple)):
                    is_mismatch_class = lambda xx: any([isinstance(ii, mm) for mm in mismatch_class])
                else:
                    is_mismatch_class = lambda xx: isinstance(ii, mismatch_class)

                method_kwarg = {} if method is None else {"method": method}  # None for using class default one
                for ii in model.layers:
                    if is_mismatch_class(ii) and ii.name in weights:
                        print(">>>> Reload layer:", ii.name)
                        ss = weights[ii.name]
                        # ss = {ww.decode().split("/")[-1] : tf.convert_to_tensor(ss[ww]) for ww in ss.attrs['weight_names']}
                        ss = {ww.decode("utf8") if hasattr(ww, "decode") else ww: np.array(ss[ww]) for ww in ss.attrs["weight_names"]}
                        ss = {kk.split("/")[-1]: vv for kk, vv in ss.items()}
                        model.get_layer(ii.name).load_resized_weights(ss, **method_kwarg)
        except Exception as error:
            print("[Error] something went wrong in load_weights_with_mismatch:", error)
            pass


def read_h5_weights(filepath):
    import h5py
    import numpy as np

    with h5py.File(filepath, "r") as h5_file:
        weights = h5_file["model_weights"] if "model_weights" in h5_file else h5_file  # full model or weights only
        return {kk: [np.array(vv[ww]) for ww in vv.attrs["weight_names"]] for kk, vv in weights.items() if len(vv) > 0}


""" Convert PyTorch weights """


def state_dict_stack_by_layer(state_dict, skip_weights=["num_batches_tracked"], unstack_weights=[]):
    stacked_state_dict = {}
    for kk, vv in state_dict.items():
        if kk[-2] == ":":
            kk = kk[:-2]  # Keras weight name like "..../weight:0"
        split_token = "/" if "/" in kk else "."
        split_kk = kk.split(split_token)
        vv = vv.numpy() if hasattr(vv, "numpy") else vv
        if split_kk[-1] in skip_weights:
            continue

        if split_kk[-1] in unstack_weights:
            stacked_state_dict[kk] = [vv]
        else:
            # split_kk[-1] in ["weight", "bias", "running_mean", "running_var", "gain"]
            layer_name = split_token.join(split_kk[:-1])
            stacked_state_dict.setdefault(layer_name, []).append(vv)
    return stacked_state_dict


def match_layer_names_with_torch(target_names, tail_align_dict={}, full_name_align_dict={}, tail_split_position=2):
    layer_names_matched_torch = [""] * len(target_names)
    raw_id_dict = {ii: id for id, ii in enumerate(target_names)}

    # is_tail_align_dict_split_by_stack = len(tail_align_dict) > 0 and isinstance(list(tail_align_dict.values())[0], dict)
    for id, ii in enumerate(target_names):
        name_split = ii.split("_")
        stack_name = name_split[0]
        head_name = "_".join(name_split[:tail_split_position])
        tail_name = "_".join(name_split[tail_split_position:])
        # cur_tail_align_dict = tail_align_dict[stack_name] if is_tail_align_dict_split_by_stack else tail_align_dict
        cur_tail_align_dict = tail_align_dict.get(stack_name, tail_align_dict)
        # print("id = {}, ii = {}, stack_name = {}, tail_name = {}".format(id, ii, stack_name, tail_name))
        if ii in full_name_align_dict:
            align = full_name_align_dict[ii]
            if isinstance(align, str):
                align = raw_id_dict[align]
            else:
                align = id + align if align < 0 else align
            layer_names_matched_torch.insert(align, ii)
            layer_names_matched_torch.pop(-1)
        elif tail_name in cur_tail_align_dict:
            align = cur_tail_align_dict[tail_name]
            if isinstance(align, str):
                align = raw_id_dict[head_name + "_" + align]
            else:
                align = id + align if align < 0 else align
            layer_names_matched_torch.insert(align, ii)
            layer_names_matched_torch.pop(-1)
        else:
            layer_names_matched_torch[id] = ii
    return layer_names_matched_torch


def align_layer_names_multi_stage(target_names, tail_align_dict={}, full_name_align_dict={}, tail_split_position=2, specific_match_func=None, verbose=1):
    if isinstance(tail_split_position, int):
        tail_align_dict, full_name_align_dict, tail_split_position = [tail_align_dict], [full_name_align_dict], [tail_split_position]
    full_name_align_dict = full_name_align_dict if isinstance(full_name_align_dict, (list, tuple)) else [full_name_align_dict]
    # for ii, jj, kk in zip(tail_align_dict, full_name_align_dict, tail_split_position):
    for idx in range(max(len(tail_split_position), len(full_name_align_dict))):
        cur_tail = tail_align_dict[idx] if idx < len(tail_align_dict) else {}
        cur_full = full_name_align_dict[idx] if idx < len(full_name_align_dict) else {}
        cur_split = tail_split_position[idx] if idx < len(tail_split_position) else tail_split_position[-1]
        if verbose > 0:
            print(">>>> tail_align_dict:", cur_tail)
            print(">>>> full_name_align_dict:", cur_full)
            print(">>>> tail_split_position:", cur_split)
        target_names = match_layer_names_with_torch(target_names, cur_tail, cur_full, cur_split)

    if specific_match_func is not None:
        target_names = specific_match_func(target_names)
    return target_names


def keras_reload_stacked_state_dict(model, stacked_state_dict, layer_names_matched_torch, additional_transfer={}, save_name=None, verbose=1):
    import numpy as np

    for kk, tf_layer_name in zip(stacked_state_dict.keys(), layer_names_matched_torch):
        if verbose > 0:
            print("  torch layer name: {}, tf layer name: {}".format(kk, tf_layer_name))
        tf_layer = model.get_layer(tf_layer_name)
        tf_weights = tf_layer.get_weights()
        torch_weight = stacked_state_dict[kk]
        if verbose > 0:
            print("    Before: [{}] torch: {}, tf: {}".format(kk, [ii.shape for ii in torch_weight], [ii.shape for ii in tf_weights]))

        if isinstance(tf_layer, layers.DepthwiseConv2D):  # DepthwiseConv2D is instance of Conv2D, put it first
            torch_weight[0] = np.transpose(torch_weight[0], (2, 3, 0, 1))
        elif isinstance(tf_layer, layers.Conv2D):
            torch_weight[0] = np.transpose(torch_weight[0], (2, 3, 1, 0))
            if len(torch_weight) > 2:  # gain
                torch_weight[2] = np.squeeze(torch_weight[2])
        elif isinstance(tf_layer, layers.PReLU):
            torch_weight[0] = np.expand_dims(np.expand_dims(torch_weight[0], 0), 0)
        elif isinstance(tf_layer, layers.Conv1D):
            torch_weight[0] = np.transpose(torch_weight[0], (2, 1, 0))
        elif isinstance(tf_layer, layers.Dense):
            # [Note] if it's fc layer after flatten, weights need to reshape according to NCHW --> NHWC
            torch_weight[0] = torch_weight[0].T

        for add_layer, add_transfer in additional_transfer.items():
            if isinstance(add_layer, str):
                if tf_layer.name.endswith(add_layer):
                    torch_weight = add_transfer(torch_weight)
            elif isinstance(tf_layer, add_layer):
                torch_weight = add_transfer(torch_weight)
        if verbose > 0:
            print("    After: [{}] torch: {}, tf: {}".format(kk, [ii.shape for ii in torch_weight], [ii.shape for ii in tf_weights]))

        tf_layer.set_weights(torch_weight)

    if save_name is None:
        save_name = model.name + ".h5"
    if len(save_name) != 0:
        if verbose > 0:
            print()
            print(">>>> Save model to:", save_name)
        model.save(save_name)


def try_save_pth_and_onnx(torch_model, save_pth=True, save_onnx=True, input_shape=(10, 3, 224, 224), dtype="float32", save_name=None):
    import torch
    import numpy as np

    save_name = torch_model.__class__.__name__ if save_name is None else save_name
    dummy_inputs = torch.from_numpy(np.random.uniform(size=input_shape).astype(dtype))
    if save_pth:
        output_name = save_name + ".pth"
        traced_cell = torch.jit.trace(torch_model, (dummy_inputs))
        torch.jit.save(traced_cell, output_name)
        print(">>>> Saved to:", output_name)

    if save_onnx:
        output_name = save_name + ".onnx"
        torch.onnx.export(
            model=torch_model,
            args=dummy_inputs,
            f=output_name,
            verbose=False,
            keep_initializers_as_inputs=True,
            training=torch.onnx.TrainingMode.PRESERVE,
            do_constant_folding=False,
            opset_version=13,
        )
        print(">>>> Saved to:", output_name)


def keras_reload_from_torch_model(
    torch_model,  # Torch model, Torch state_dict wights or Torch model weights file
    keras_model=None,
    input_shape=(224, 224),
    skip_weights=["num_batches_tracked"],
    unstack_weights=[],
    tail_align_dict={},
    full_name_align_dict={},
    tail_split_position=2,
    additional_transfer={},
    specific_match_func=None,
    save_name=None,
    do_convert=True,
    do_predict=True,
    verbose=1,
):
    import torch
    import numpy as np
    import tensorflow as tf
    from tensorflow import keras
    from keras_cv_attention_models import test_images

    input_shape = input_shape[:2] if keras_model is None else keras_model.input_shape[1:-1]
    if isinstance(torch_model, str):
        print(">>>> Reload Torch weight file:", torch_model)
        torch_model = torch.load(torch_model, map_location=torch.device("cpu"))
        torch_model = torch_model.get("model", torch_model.get("state_dict", torch_model))
    is_state_dict = isinstance(torch_model, dict)

    """ Chelsea the cat  """
    if do_predict:
        img = test_images.cat()
        img = keras.applications.imagenet_utils.preprocess_input(tf.image.resize(img, input_shape), mode="torch").numpy()

    if not is_state_dict:
        _ = torch_model.eval()
        state_dict = torch_model.state_dict()

        if do_predict:
            try:
                # from torchsummary import summary
                # summary(torch_model, (3, *input_shape), device="cpu")
                """Torch Run predict"""
                out = torch_model(torch.from_numpy(np.expand_dims(img.transpose(2, 0, 1), 0).astype("float32")))
                out = out.detach().cpu().numpy()
                # out = tf.nn.softmax(out).numpy()  # If classifier activation is not softmax
                torch_out = keras.applications.imagenet_utils.decode_predictions(out)
            except:
                pass
    else:
        state_dict = torch_model

    """ Convert torch weights """
    # torch_params = {kk: (np.cumproduct(vv.shape)[-1] if len(vv.shape) != 0 else 1) for kk, vv in state_dict.items() if ".num_batches_tracked" not in kk}
    stacked_state_dict = state_dict_stack_by_layer(state_dict, skip_weights=skip_weights, unstack_weights=unstack_weights)
    if verbose > 0:
        print(">>>> torch_model total_parameters :", np.sum([np.sum([np.prod(jj.shape) for jj in ii]) for ii in stacked_state_dict.values()]))
    aa = {kk: [1 if isinstance(jj, float) else jj.shape for jj in vv] for kk, vv in stacked_state_dict.items()}
    if verbose > 0:
        print(">>>> Torch weights:")
        _ = [print("  '{}': {}".format(kk, vv)) for kk, vv in aa.items()]
        print()

    if keras_model is None:
        return

    """ Keras model weights """
    target_names = [ii.name for ii in keras_model.layers if len(ii.weights) != 0]
    aa = {keras_model.get_layer(ii).name: [jj.shape.as_list() for jj in keras_model.get_layer(ii).weights] for ii in target_names}
    if verbose > 0:
        print(">>>> keras_model total_parameters :", np.sum([np.sum([int(np.prod(jj)) for jj in ii]) for ii in aa.values()]))
        print(">>>> Keras weights:")
        _ = [print("  '{}': {}".format(kk, vv)) for kk, vv in aa.items()]
        print()

    """ Load torch weights and save h5 """
    if len(tail_align_dict) != 0 or len(full_name_align_dict) != 0 or specific_match_func is not None:
        aligned_names = align_layer_names_multi_stage(target_names, tail_align_dict, full_name_align_dict, tail_split_position, specific_match_func, verbose)
        aa = {keras_model.get_layer(ii).name: [jj.shape.as_list() for jj in keras_model.get_layer(ii).weights] for ii in aligned_names}
        if verbose > 0:
            print(">>>> Keras weights matched torch:")
            _ = [print("  '{}': {}".format(kk, vv)) for kk, vv in aa.items()]
            print()
    else:
        aligned_names = target_names

    if not do_convert:
        return

    save_name = save_name if save_name is not None else keras_model.name + "_imagenet.h5"
    if verbose > 0:
        print(">>>> Keras reload torch weights:")
    keras_reload_stacked_state_dict(keras_model, stacked_state_dict, aligned_names, additional_transfer, save_name=save_name, verbose=verbose)
    if verbose > 0:
        print()

    """ Keras run predict """
    if do_predict:
        try:
            pred = keras_model(tf.expand_dims(img, 0)).numpy()
            # pred = tf.nn.softmax(pred).numpy()  # If classifier activation is not softmax
            print(">>>> Keras model prediction:", keras.applications.imagenet_utils.decode_predictions(pred)[0])
            print()
            if not is_state_dict:
                print(">>>> Torch model prediction:", torch_out)
        except:
            pass



import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format, initializers
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    BiasLayer,
    ChannelAffine,
    ClassToken,
    conv2d_no_bias,
    drop_block,
    drop_connect_rates_split,
    PositionalEmbedding,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

LAYER_NORM_EPSILON = 1e-6

PRETRAINED_DICT = {
    "beit_base_patch16": {"imagenet21k-ft1k": {224: "d7102337a13a3983f3b6470de77b5d5c", 384: "76353026477c60f8fdcbcc749fea17b3"}},
    "beit_v2_base_patch16": {"imagenet21k-ft1k": {224: "d001dcb67cdda16bfdbb2873ab9b13c8"}},
    "beit_large_patch16": {
        "imagenet21k-ft1k": {224: "fce2d162e7fa4dba9a1b1fc5e1dec5ce", 384: "158934d07dd8b1e1c6b96883aa00a748", 512: "64d18088e91df243960e5830aab80a6e"}
    },
    "beit_v2_large_patch16": {"imagenet21k-ft1k": {224: "b3cee12a545bfb676f9f426ee7158d27"}},
    "eva_giant_patch14": {
        "imagenet21k-ft1k": {224: "5a475db6696d6e36ea896ec5dbd1c20d", 336: "fd8eeec10d6b6cb607ce033ea85b8e80", 560: "0ef0d2961523fb2047fbdb59cc347c17"}
    },
    "eva_large_patch14": {"imagenet21k-ft1k": {196: "bbeea886fbde4bd1c8c9876345273a99", 336: "4928faafd0177fe8f0d02dab4abc8e83"}},
    "flexivit_small": {"imagenet": {240: "efb73a97d099a491b69ebfaf8a337df8"}},
    "flexivit_base": {"imagenet": {240: "dac627debb194928db01e1b9b7a548fd"}},
    "flexivit_large": {"imagenet": {240: "6faa953227d2ef1df6758f8eb7234490"}},
}


@backend.register_keras_serializable(package="beit")
class MultiHeadRelativePositionalEmbedding(layers.Layer):
    def __init__(self, with_cls_token=True, attn_height=-1, num_heads=-1, **kwargs):
        super().__init__(**kwargs)
        self.with_cls_token, self.attn_height, self.num_heads = with_cls_token, attn_height, num_heads
        if with_cls_token:
            self.cls_token_len = 1
            self.cls_token_pos_len = 3
        else:
            self.cls_token_len = 0
            self.cls_token_pos_len = 0

    def build(self, attn_shape):
        # input (with_cls_token=True): `[batch, num_heads, attn_blocks, attn_blocks]`. where `attn_blocks = attn_height * attn_width + class_token`
        # input (with_cls_token=False): `[batch, num_heads, attn_blocks, attn_blocks]`. where `attn_blocks = attn_height * attn_width`
        # print(attn_shape)
        if self.attn_height == -1:
            height = width = int(float(attn_shape[2] - self.cls_token_len) ** 0.5)  # hh == ww, e.g. 14
        else:
            height = self.attn_height
            width = int(float(attn_shape[2] - self.cls_token_len) / height)
        num_heads = attn_shape[1] if self.num_heads == -1 else self.num_heads
        num_relative_distance = (2 * height - 1) * (2 * width - 1) + self.cls_token_pos_len
        # pos_shape = (num_relative_distance, num_heads)
        pos_shape = (num_heads, num_relative_distance)
        # initializer = tf.random_normal_initializer(stddev=0.02)
        self.relative_position_bias_table = self.add_weight(name="positional_embedding", shape=pos_shape, initializer="zeros", trainable=True)

        hh, ww = np.meshgrid(range(height), range(width))  # tf.meshgrid is same with np.meshgrid 'xy' mode, while torch.meshgrid 'ij' mode
        coords = np.stack([hh, ww], axis=-1)  # [14, 14, 2]
        coords_flatten = np.reshape(coords, [-1, 2])  # [196, 2]
        relative_coords = coords_flatten[:, None, :] - coords_flatten[None, :, :]  # [196, 196, 2]
        relative_coords_hh = relative_coords[:, :, 0] + height - 1
        relative_coords_ww = (relative_coords[:, :, 1] + width - 1) * (2 * height - 1)
        relative_coords = np.stack([relative_coords_hh, relative_coords_ww], axis=-1)

        relative_position_index = np.sum(relative_coords, axis=-1).astype("float32")  # [196, 196]
        if attn_shape[3] != attn_shape[2]:
            # Choose the small values if value_block != query_block
            relative_position_index = relative_position_index[:, -(attn_shape[3] - self.cls_token_len) :]

        if self.with_cls_token:
            top = np.ones((1, relative_position_index.shape[1]), dtype=relative_position_index.dtype) * (num_relative_distance - 3)
            left = np.ones((relative_position_index.shape[0], 1), dtype=relative_position_index.dtype) * (num_relative_distance - 2)
            corner = np.ones((1, 1), dtype=relative_position_index.dtype) * (num_relative_distance - 1)
            # print(f">>>> {top.shape = }, {left.shape = }, {corner.shape = }")
            # >>>> top.shape = TensorShape([1, 196]), left.shape = TensorShape([196, 1]), corner.shape = TensorShape([1, 1])
            left_corner = np.concatenate([corner, left], axis=0)
            relative_position_index = np.concatenate([top, relative_position_index], axis=0)
            relative_position_index = np.concatenate([left_corner, relative_position_index], axis=1)  # [197, 197]
        self.relative_position_index = functional.convert_to_tensor(relative_position_index, dtype="int64")
        super().build(attn_shape)

    def call(self, attention_scores, **kwargs):
        pos_emb = functional.gather(self.relative_position_bias_table, self.relative_position_index, axis=1)
        # tf.print(pos_emb.shape, attention_scores.shape)
        return attention_scores + pos_emb

    def get_config(self):
        base_config = super().get_config()
        base_config.update({"with_cls_token": self.with_cls_token, "attn_height": self.attn_height, "num_heads": self.num_heads})
        return base_config

    def load_resized_weights(self, source_layer, method="bilinear"):
        if isinstance(source_layer, dict):
            source_tt = list(source_layer.values())[0]  # weights
            # source_tt = source_layer["pos_emb:0"]  # weights
        else:
            source_tt = source_layer.get_weights()[0]  # layer
        source_tt = np.array(source_tt.detach() if hasattr(source_tt, "detach") else source_tt).astype("float32")
        # self.relative_position_bias_table.assign(tf.transpose(source_tt))
        hh = ww = int(float(source_tt.shape[1] - self.cls_token_pos_len) ** 0.5)  # assume source weights are all square shape
        num_heads = source_tt.shape[0]
        ss = source_tt[:, : hh * ww].reshape((num_heads, hh, ww))  # [num_heads, hh, ww]

        if self.attn_height == -1:
            target_hh = target_ww = int(float(self.relative_position_bias_table.shape[1] - self.cls_token_pos_len) ** 0.5)
        else:
            target_hh = 2 * self.attn_height - 1
            target_ww = int(float(self.relative_position_bias_table.shape[1] - self.cls_token_pos_len) / target_hh)

        tt = backend.numpy_image_resize(ss, target_shape=[target_hh, target_ww], method=method, is_source_channels_last=False)
        tt = tt.reshape((num_heads, tt.shape[1] * tt.shape[2]))  # [num_heads, target_hh * target_ww]
        if self.with_cls_token:
            tt = np.concatenate([tt, source_tt[:, -self.cls_token_pos_len :]], axis=1)
        self.set_weights([tt])

    def show_pos_emb(self, rows=1, base_size=2):
        import math
        import matplotlib.pyplot as plt

        num_heads = self.relative_position_bias_table.shape[0]
        # pos_emb = tf.gather(self.relative_position_bias_table, self.relative_position_index, axis=1).numpy()
        hh = ww = int(float(self.relative_position_bias_table.shape[1] - self.cls_token_pos_len) ** 0.5)
        pos_emb = self.relative_position_bias_table[:, : hh * ww]
        pos_emb = pos_emb.detach().numpy() if hasattr(pos_emb, "detach") else pos_emb.numpy()
        pos_emb = pos_emb.reshape((num_heads, hh, ww))
        cols = int(math.ceil(num_heads / rows))
        fig, axes = plt.subplots(rows, cols, figsize=(base_size * cols, base_size * rows))
        for id, ax in enumerate(axes.flatten()):
            if id >= num_heads:
                break
            ax.imshow(pos_emb[id])
            ax.set_axis_off()
        fig.tight_layout()
        return fig


def scaled_dot_product_attention(query, key, value, output_shape, pos_emb=None, out_weight=True, out_bias=False, qk_scale=-1, dropout=0, name=None):
    output_dim = output_shape[-1]
    blocks = output_shape[1:-1] if output_shape[0] is None or output_shape[0] < 1 else output_shape[:-1]
    # query, value: [batch, num_heads, blocks, key_dim], key: [batch, num_heads, key_dim, blocks]
    qk_scale = qk_scale if qk_scale > 0 else (1.0 / (float(query.shape[-1]) ** 0.5))
    # print(f"{query.shape = }, {key.shape = }")
    # attention_scores = layers.Lambda(lambda xx: functional.matmul(xx[0], xx[1]))([query, key]) * qk_scale  # [batch, num_heads, q_blocks, k_blocks]
    attention_scores = query @ key
    if qk_scale != 1:
        attention_scores = attention_scores * qk_scale
    # print(f"{attention_scores.shape = }")
    if pos_emb is not None:
        # attention_scores = MultiHeadPositionalEmbedding(query_height=height, name=name and name + "attn_pos")(attention_scores)
        attention_scores = pos_emb(attention_scores)
    attention_scores = layers.Softmax(axis=-1, name=name and name + "attention_scores")(attention_scores)
    if dropout > 0:
        attention_scores = layers.Dropout(dropout, name=name and name + "attn_drop")(attention_scores)

    # output = functional.matmul(attention_scores, value)    # [batch, num_heads, q_blocks, key_dim * attn_ratio]
    # output = layers.Lambda(lambda xx: functional.matmul(xx[0], xx[1]))([attention_scores, value])
    attention_output = attention_scores @ value
    output = functional.transpose(attention_output, perm=[0, 2, 1, 3])  # [batch, q_blocks, num_heads, key_dim * attn_ratio]
    output = functional.reshape(output, [-1, *blocks, output.shape[2] * output.shape[3]])  # [batch, q_blocks, channel * attn_ratio]

    if out_weight:
        # [batch, hh, ww, num_heads * key_dim] * [num_heads * key_dim, out] --> [batch, hh, ww, out]
        output_dim = output_dim if output_dim > 0 else value.shape[-1]
        output = layers.Dense(output_dim, use_bias=out_bias, name=name and name + "output")(output)
    return output


def qkv_to_multi_head_channels_last_format(query, key, value, num_heads, data_format=None):
    data_format = image_data_format() if data_format is None else data_format
    # print(f">>>> {query.shape = }, {key.shape = }, {value.shape = }, {num_heads = }, {data_format = }")
    if data_format == "channels_last":
        if query is not None:
            # query [batch, hh, ww, channel] -> [batch, num_heads, hh * ww, key_dim]
            query = functional.reshape(query, [-1, np.prod(query.shape[1:-1]), num_heads, query.shape[-1] // num_heads])
            query = functional.transpose(query, [0, 2, 1, 3])
        if key is not None:
            # key [batch, hh, ww, channel] -> [batch, num_heads, key_dim, hh * ww]
            key = functional.reshape(key, [-1, np.prod(key.shape[1:-1]), num_heads, key.shape[-1] // num_heads])
            key = functional.transpose(key, [0, 2, 3, 1])
        if value is not None:
            # value [batch, hh, ww, channel] -> [batch, num_heads, hh * ww, vv_dim]
            value = functional.reshape(value, [-1, np.prod(value.shape[1:-1]), num_heads, value.shape[-1] // num_heads])
            value = functional.transpose(value, [0, 2, 1, 3])
    else:
        if query is not None:
            # query [batch, channel, hh, ww] -> [batch, num_heads, hh * ww, key_dim]
            query = functional.reshape(query, [-1, num_heads, query.shape[1] // num_heads, np.prod(query.shape[2:])])
            query = functional.transpose(query, [0, 1, 3, 2])
        if key is not None:
            # key [batch, channel, hh, ww] -> [batch, num_heads, key_dim, hh * ww]
            key = functional.reshape(key, [-1, num_heads, key.shape[1] // num_heads, np.prod(key.shape[2:])])
        if value is not None:
            # value [batch, channel, hh, ww] -> [batch, num_heads, hh * ww, vv_dim]
            value = functional.reshape(value, [-1, num_heads, value.shape[1] // num_heads, np.prod(value.shape[2:])])
            value = functional.transpose(value, [0, 1, 3, 2])
    # print(f">>>> {query.shape = }, {key.shape = }, {value.shape = }, {num_heads = }, {data_format = }")
    return query, key, value


def attention_block(
    inputs, num_heads=4, key_dim=0, qv_bias=True, qkv_bias=False, out_weight=True, out_bias=False, use_pos_emb=False, attn_height=-1, attn_dropout=0, name=None
):
    _, bb, cc = inputs.shape
    key_dim = key_dim if key_dim > 0 else cc // num_heads
    emded_dim = num_heads * key_dim
    # print(f">>>> {bb = }, {cc = }, {emded_dim = }")

    # if qkv_bias, just use bias in qkv_dense, and set qv_bias False
    qkv_bias, qv_bias = (True, False) if qkv_bias else (False, qv_bias)

    qkv = layers.Dense(emded_dim * 3, use_bias=qkv_bias, name=name and name + "qkv")(inputs)
    # print(f">>>> {qkv.shape = }")
    qkv = functional.reshape(qkv, [-1, bb, qkv.shape[-1]])
    query, key, value = functional.split(qkv, 3, axis=-1)
    # query = [batch, num_heads, cls_token + hh * ww, key_dim]
    if qv_bias:
        query = BiasLayer(name=name + "query_bias")(query)
    if qv_bias:
        value = BiasLayer(name=name + "value_bias")(value)
    query, key, value = qkv_to_multi_head_channels_last_format(query, key, value, num_heads, data_format="channels_last")

    pos_emb = MultiHeadRelativePositionalEmbedding(attn_height=attn_height, name=name and name + "pos_emb") if use_pos_emb else None
    output_shape = [-1, bb, emded_dim]
    return scaled_dot_product_attention(query, key, value, output_shape, pos_emb, out_weight, out_bias, dropout=attn_dropout, name=name)


def attention_mlp_block(inputs, embed_dim, gamma_init_value=0.1, mlp_ratio=4, drop_rate=0, activation="gelu", attn_params={}, name=""):
    # print(f">>>> {inputs.shape = }, {drop_rate = }")
    nn = layers.LayerNormalization(axis=-1, epsilon=LAYER_NORM_EPSILON, name=name + "attn_ln")(inputs)  # "channels_first" also using axis=-1
    nn = attention_block(nn, **attn_params, name=name + "attn_")
    nn = ChannelAffine(use_bias=False, weight_init_value=gamma_init_value, name=name + "attn_gamma")(nn) if gamma_init_value > 0 else nn
    nn = drop_block(nn, drop_rate)
    attn_out = layers.Add(name=name + "attn_out")([inputs, nn])

    """ MLP """
    nn = layers.LayerNormalization(axis=-1, epsilon=LAYER_NORM_EPSILON, name=name + "mlp_ln")(attn_out)  # "channels_first" also using axis=-1
    nn = layers.Dense(embed_dim * mlp_ratio, name=name + "mlp_dense_1")(nn)
    nn = activation_by_name(nn, activation, name=name + "mlp_")
    nn = layers.Dense(embed_dim, name=name + "mlp_dense_2")(nn)
    nn = ChannelAffine(use_bias=False, weight_init_value=gamma_init_value, name=name + "mlp_gamma")(nn) if gamma_init_value > 0 else nn
    nn = drop_block(nn, drop_rate)
    # print(f">>>> {attn_out.shape = }, {nn.shape = }")
    nn = layers.Add(name=name + "mlp_output")([attn_out, nn])
    return nn


@backend.register_keras_serializable(package="beit")
class HeadInitializer(initializers.Initializer):
    def __init__(self, stddev=0.02, scale=0.001, **kwargs):
        super().__init__(**kwargs)
        self.stddev, self.scale = stddev, scale

    def __call__(self, shape, dtype="float32"):
        return initializers.TruncatedNormal(stddev=self.stddev)(shape, dtype=dtype) * self.scale

    def get_config(self):
        base_config = super().get_config()
        base_config.update({"stddev": self.stddev, "scale": self.scale})
        return base_config


@backend.register_keras_serializable(package="beit")
class PatchConv2DWithResampleWeights(layers.Conv2D):
    def load_resized_weights(self, source_layer, method="bilinear"):
        import numpy as np

        # print("source_layer:", {kk: vv.shape for kk, vv in source_layer.items()})
        if isinstance(source_layer, dict):
            source_kernel, source_bias = list(source_layer.values())  # weights
        else:
            source_kernel, source_bias = source_layer.get_weights()  # layer

        # From FlexiViT https://github.com/google-research/big_vision/blob/main/big_vision/models/proj/flexi/vit.py#L30
        # Paper [PDF 2212.08013 FlexiViT: One Model for All Patch Sizes](https://arxiv.org/pdf/2212.08013.pdf)

        # channels_last source_kernel shape `[patch_size, patch_size, in_channel, out_channel]`
        # channels_first source_kernel shape `[out_channel, in_channel, patch_size, patch_size]`
        source_kernel, source_bias = np.array(source_kernel).astype("float32"), np.array(source_bias).astype("float32")
        source_shape, target_shape = source_kernel.shape[:2], self.kernel_size  # source_kernel is from h5, must be channels_last format
        # print(f"{source_shape = }, {target_shape = }")

        # get_resize_mat(old_shape, target_shape)
        # NOTE: we are using tf.image.resize here to match the resize operations in
        # the data preprocessing pipeline.
        mat = []
        for idx in range(source_shape[0] * source_shape[1]):
            basis_vec = np.zeros(source_shape).astype("float32")
            basis_vec[np.unravel_index(idx, source_shape)] = 1.0
            vec = np.expand_dims(basis_vec, -1 if image_data_format() == "channels_last" else 0)
            vec = functional.resize(vec, target_shape, method=method)
            vec = vec.detach().numpy() if hasattr(vec, "detach") else vec.numpy()
            mat.append(vec.reshape(-1))
        resize_mat_pinv = np.linalg.pinv(np.stack(mat))

        # v_resample_kernel = jax.vmap(jax.vmap(lambda kernel: (resize_mat_pinv @ kernel.reshape(-1)).reshape(new_hw), 2, 2), 3, 3)
        # cc = v_resample_kernel(old)
        # As it's only one weight, just using two loop here, instead of `jax.vmap`
        target_weights = np.stack([[(resize_mat_pinv @ jj.reshape(-1)).reshape(target_shape) for jj in ii] for ii in source_kernel.transpose([3, 2, 0, 1])])
        if image_data_format() == "channels_last":
            target_weights = target_weights.transpose([2, 3, 1, 0])
        self.set_weights([target_weights, source_bias])


def Beit(
    depth=12,
    embed_dim=768,
    num_heads=12,
    mlp_ratio=4,
    patch_size=16,
    attn_key_dim=0,
    attn_qv_bias=True,  # Default False for Vit, True for Beit, if True and attn_qkv_bias being False, will add BiasLayer for query and key.
    attn_qkv_bias=False,  # True for Vit, False for Beit, if True, will just use bias in qkv_dense, and set qv_bias False.
    attn_out_weight=True,
    attn_out_bias=True,
    attn_dropout=0,
    gamma_init_value=0.1,  # 0 for Vit, 0.1 for Beit, if > 0 will use `layer_scale` on block output
    use_abs_pos_emb=False,  # True for Vit, False for Beit, whether use abcolute positional embedding or relative one in attention blocks
    use_abs_pos_emb_on_cls_token=True,  # False for FlexiViT, no_embed_class in timm. If use_abs_pos_emb is True, whether apply pos_emb on cls_token.
    use_mean_pooling=True,  # False for Vit, True for Beit, whether use use mean output or `class_token` output
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="gelu",
    drop_connect_rate=0,
    classifier_activation="softmax",
    pretrained=None,
    force_reload_mismatch=False,  # set True if patch_size changed, will force reloading pos_emb and stem_conv weights
    model_name="beit",
    kwargs=None,
):
    # Regard input_shape as force using original shape if first element is None or -1,
    # else assume channel dimention is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)

    """ forward_embeddings """
    # nn = conv2d_no_bias(inputs, embed_dim, patch_size, strides=patch_size, padding="valid", use_bias=True, name="stem_")
    nn = PatchConv2DWithResampleWeights(embed_dim, patch_size, strides=patch_size, padding="valid", use_bias=True, name="stem_conv")(inputs)
    nn = nn if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(nn)  # channels_first -> channels_last
    patch_height = nn.shape[1]
    nn = layers.Reshape([-1, nn.shape[-1]])(nn)

    if use_abs_pos_emb and use_abs_pos_emb_on_cls_token:  # EvaLarge and EvaGiant
        nn = ClassToken(name="cls_token")(nn)
        nn = PositionalEmbedding(input_height=patch_height, name="positional_embedding")(nn)
    elif use_abs_pos_emb:  # FlexiViT models
        nn = PositionalEmbedding(input_height=patch_height, name="positional_embedding")(nn)
        nn = ClassToken(name="cls_token")(nn)
    else:  # Beit and BeitV2
        nn = ClassToken(name="cls_token")(nn)

    attn_params = {
        "num_heads": num_heads,
        "key_dim": attn_key_dim,
        "qv_bias": attn_qv_bias,
        "qkv_bias": attn_qkv_bias,
        "out_weight": attn_out_weight,
        "out_bias": attn_out_bias,
        "attn_height": patch_height,
        "attn_dropout": attn_dropout,
        "use_pos_emb": not use_abs_pos_emb,
    }

    """ forward_tokens """
    drop_connect_rates = drop_connect_rates_split([depth], 0.0, drop_connect_rate)[0]
    for id in range(depth):
        name = "block{}_".format(id)
        block_drop_rate = drop_connect_rates[id]
        nn = attention_mlp_block(nn, embed_dim, gamma_init_value, mlp_ratio, block_drop_rate, activation, attn_params, name=name)

    if use_mean_pooling:
        nn = functional.reduce_mean(nn[:, 1:, :], axis=1)
        nn = layers.LayerNormalization(axis=-1, epsilon=LAYER_NORM_EPSILON, name="out_ln")(nn)  # "channels_first" also using axis=-1
    else:
        nn = layers.LayerNormalization(axis=-1, epsilon=LAYER_NORM_EPSILON, name="out_ln")(nn)  # "channels_first" also using axis=-1
        nn = nn[:, 0]

    if num_classes > 0:
        head_init = HeadInitializer()
        nn = layers.Dense(
            num_classes, dtype="float32", activation=classifier_activation, kernel_initializer=head_init, bias_initializer=head_init, name="predictions"
        )(nn)
    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="tf")
    mismatch_class = [PatchConv2DWithResampleWeights, PositionalEmbedding if use_abs_pos_emb else MultiHeadRelativePositionalEmbedding]
    reload_model_weights(model, PRETRAINED_DICT, "beit", pretrained, mismatch_class, force_reload_mismatch)
    return model


def BeitBasePatch16(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet21k-ft1k", **kwargs):
    embed_dim = 768
    depth = 12
    num_heads = 12
    gamma_init_value = 0.1
    force_reload_mismatch = kwargs.get("patch_size", 16) != 16  # If patch_size not 16, force reload pos_emb and stem_conv weights
    kwargs.pop("kwargs", None)  # From BeitV2BasePatch16
    return Beit(**locals(), model_name=kwargs.pop("model_name", "beit_base_patch16"), **kwargs)


def BeitV2BasePatch16(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet21k-ft1k", **kwargs):
    return BeitBasePatch16(**locals(), **kwargs, model_name="beit_v2_base_patch16")


def BeitLargePatch16(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet21k-ft1k", **kwargs):
    embed_dim = 1024
    depth = 24
    num_heads = 16
    gamma_init_value = 1e-5
    force_reload_mismatch = kwargs.get("patch_size", 16) != 16  # If patch_size not 16, force reload pos_emb and stem_conv weights
    kwargs.pop("kwargs", None)  # From BeitV2LargePatch16
    return Beit(**locals(), model_name=kwargs.pop("model_name", "beit_large_patch16"), **kwargs)


def BeitV2LargePatch16(
    input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet21k-ft1k", **kwargs
):
    return BeitLargePatch16(**locals(), **kwargs, model_name="beit_v2_large_patch16")


""" keras_model_load_weights_from_pytorch_model """


def keras_model_load_weights_from_pytorch_model(keras_model, timm_vit_model, save_name=None):
    from keras_cv_attention_models import download_and_load, attention_layers

    skip_weights = ["relative_position_index"]
    unstack_weights = ["cls_token", "gamma_1", "gamma_2", "relative_position_bias_table", "q_bias", "v_bias", "pos_embed"]
    tail_align_dict = {"attn_gamma": -6, "mlp_gamma": -9, "attn_query_bias": -1, "attn_value_bias": -1, "attn_pos_emb": -1}
    full_name_align_dict = {"cls_token": -2 if "flexivit" in keras_model.name else -1, "positional_embedding": -1}
    additional_transfer = {attention_layers.MultiHeadRelativePositionalEmbedding: lambda ww: [ww[0].T]}

    download_and_load.keras_reload_from_torch_model(
        torch_model=timm_vit_model,
        keras_model=keras_model,
        input_shape=keras_model.input_shape[1:-1],
        skip_weights=skip_weights,
        unstack_weights=unstack_weights,
        tail_align_dict=tail_align_dict,
        full_name_align_dict=full_name_align_dict,
        tail_split_position=1,
        additional_transfer=additional_transfer,
        save_name=save_name if save_name is not None else (keras_model.name + "_{}.h5".format(keras_model.input_shape[1])),
        do_convert=True,
        # do_predict=False if "eva_giant" in keras_model.name else True,
    )



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    anti_alias_downsample,
    batchnorm_with_activation,
    conv2d_no_bias,
    drop_block,
    drop_connect_rates_split,
    se_module,
    eca_module,
    # output_block,
    add_pre_post_process,
)
from keras_cv_attention_models import attention_layers

# from keras_cv_attention_models.download_and_load import reload_model_weights

DEFAULT_PARAMS = {
    "bot": {"num_heads": 4, "relative": True, "out_bias": False},
    "halo": {"num_heads": 8, "block_size": 4, "halo_size": 1},
    "sa": {"kernel_size": 3, "groups": 2, "downsample_first": False},
    "cot": {"kernel_size": 3, "downsample_first": True},
    "outlook": {"num_heads": 8, "kernel_size": 3},
}


def attn_block(
    inputs,
    filters,
    strides=1,
    attn_type=None,
    attn_params={},
    se_ratio=0,
    use_eca=False,
    groups=1,
    group_size=0,
    bn_after_attn=True,
    use_evo_norm=False,
    evo_norm_group_size=-1,
    epsilon=1e-5,
    activation="relu",
    name="",
):
    nn = inputs
    if attn_params is not None and len(attn_params) != 0:
        default_attn_params = DEFAULT_PARAMS.get(attn_type, {}).copy()
        default_attn_params.update(attn_params)
        attn_params = default_attn_params
    else:
        attn_params = DEFAULT_PARAMS.get(attn_type, {})

    attn_act = attn_params.pop("activation", activation)
    se_divisor = attn_params.pop("se_divisor", 8)
    attn_block = None if attn_type is None else getattr(attention_layers, attn_type, None)
    need_downsaple = False
    if attn_block is not None:
        nn = attn_block(nn, **attn_params, name=name + "attn_")
        need_downsaple = True
    elif attn_type == "bot":  # mhsa_with_relative_position_embedding from botnet
        nn = attention_layers.mhsa_with_relative_position_embedding(nn, **attn_params, name=name + "mhsa_")
        need_downsaple = True
    elif attn_type == "halo":  # halo_attention from halonet
        halo_expansion = attn_params.pop("halo_expansion", 1)
        out_shape = int(filters * halo_expansion)
        nn = attention_layers.halo_attention(nn, **attn_params, strides=strides, out_shape=out_shape, name=name + "halo")
    elif attn_type == "sa":  # split_attention_conv2d from resnest
        nn = attention_layers.split_attention_conv2d(nn, **attn_params, filters=filters, strides=strides, activation=attn_act, name=name + "sa_")
    elif attn_type == "cot":  # cot_attention from cotnet
        nn = attention_layers.cot_attention(nn, **attn_params, strides=strides, activation=attn_act, name=name + "cot_")
    elif attn_type == "outlook":  # outlook_attention from volo
        nn = attention_layers.outlook_attention(nn, filters, **attn_params, name=name + "outlook_")
        need_downsaple = True
    # elif attn_type == "groups_conv":  # ResNeXt like
    #     nn = conv2d_no_bias(nn, filters, **attn_params, strides=strides, padding="SAME", name=name + "GC_")
    else:  # ResNet and `groups > 1` for ResNeXt like
        groups = groups if group_size == 0 else filters // group_size
        conv_name = (name + "GC_") if groups > 1 else name
        nn = conv2d_no_bias(nn, filters, 3, strides=strides, padding="SAME", groups=groups, name=conv_name)

    # if strides != 1 and (nn.shape[1] == inputs.shape[1] or nn.shape[2] == inputs.shape[2]):  # Downsample
    if strides != 1 and need_downsaple:  # Downsample
        # nn = layers.ZeroPadding2D(padding=1, name=name + "pad")(nn)
        nn = layers.AvgPool2D(pool_size=2, strides=strides, name=name + "pool")(nn)

    if bn_after_attn:
        bn_params = {"use_evo_norm": use_evo_norm, "evo_norm_group_size": evo_norm_group_size, "epsilon": epsilon}
        nn = batchnorm_with_activation(nn, activation, zero_gamma=False, **bn_params, name=name)

    if attn_type is None and se_ratio:
        nn = se_module(nn, se_ratio=se_ratio, divisor=se_divisor, activation=attn_act, name=name + "se_")

    if attn_type is None and use_eca:
        nn = eca_module(nn, name=name + "eca_")
    return nn


def conv_shortcut_branch(inputs, filters, preact=False, strides=1, shortcut_type="conv", bn_act_params={}, name=""):
    if shortcut_type is None:
        return None

    if strides > 1 and shortcut_type == "avg":
        shortcut = layers.AvgPool2D(strides, strides=strides, padding="SAME", name=name + "shortcut_down")(inputs)
        strides = 1
    elif strides > 1 and shortcut_type == "anti_alias":
        shortcut = anti_alias_downsample(inputs, kernel_size=3, strides=2, name=name + "shortcut_down")
        strides = 1
    else:
        shortcut = inputs
    shortcut = conv2d_no_bias(shortcut, filters, 1, strides=strides, name=name + "shortcut_")
    if not preact:  # ResNet
        shortcut = batchnorm_with_activation(shortcut, zero_gamma=False, **bn_act_params, name=name + "shortcut_")
    return shortcut


def deep_branch(
    inputs, filters, strides=1, hidden_channel_ratio=0.25, use_3x3_kernel=False, bn_after_attn=True, bn_act_params={}, attn_block_params={}, name=""
):
    hidden_filter = int(filters * hidden_channel_ratio)
    if use_3x3_kernel:
        nn = conv2d_no_bias(inputs, hidden_filter, 3, strides=1, padding="SAME", name=name + "deep_1_")  # Using strides=1 for not changing input shape
        # nn = conv2d_no_bias(inputs, hidden_filter, 3, strides=strides, padding="SAME", name=name + "1_")
        # strides = 1
    else:
        nn = conv2d_no_bias(inputs, hidden_filter, 1, strides=1, padding="VALID", name=name + "deep_1_")
    nn = batchnorm_with_activation(nn, zero_gamma=False, **bn_act_params, name=name + "deep_1_")
    # bn_after_attn = False if use_3x3_kernel else bn_after_attn
    nn = attn_block(nn, hidden_filter, strides, **attn_block_params, **bn_act_params, bn_after_attn=bn_after_attn, name=name + "deep_2_")

    if not use_3x3_kernel:
        nn = conv2d_no_bias(nn, filters, 1, strides=1, padding="VALID", name=name + "deep_3_")
    return nn


def aot_block(
    inputs,
    filters,
    strides=1,
    conv_shortcut=False,
    hidden_channel_ratio=0.25,
    drop_rate=0,
    preact=False,
    use_3x3_kernel=False,
    bn_after_attn=True,
    shortcut_type="conv",
    use_block_output_activation=True,
    use_evo_norm=False,
    epsilon=1e-5,
    evo_norm_group_size=-1,
    activation="relu",
    attn_block_params={},
    name="",
):
    if attn_block_params.get("attn_type", None) == "halo":  # HaloAttention
        halo_block_size = attn_block_params.get("attn_params", {}).get("block_size", DEFAULT_PARAMS["halo"]["block_size"])
        height, width = inputs.shape[1:-1] if image_data_format() == "channels_last" else inputs.shape[2:]
        if height % halo_block_size != 0 or width % halo_block_size != 0:
            gap_h = halo_block_size - height % halo_block_size
            gap_w = halo_block_size - width % halo_block_size
            pad_head_h, pad_tail_h = gap_h // 2, gap_h - gap_h // 2
            pad_head_w, pad_tail_w = gap_w // 2, gap_w - gap_w // 2
            # print(f">>>> Halo pad: {inputs.shape = }, {pad_head_h = }, {pad_tail_h = }, {pad_head_w = }, {pad_tail_w = }")
            inputs = layers.ZeroPadding2D(padding=((pad_head_h, pad_tail_h), (pad_head_w, pad_tail_w)), name=name + "gap_pad")(inputs)

    bn_params = {"use_evo_norm": use_evo_norm, "evo_norm_group_size": evo_norm_group_size, "epsilon": epsilon}
    bn_act_params = {"activation": activation, "use_evo_norm": use_evo_norm, "evo_norm_group_size": evo_norm_group_size, "epsilon": epsilon}
    if preact:  # ResNetV2
        pre_inputs = batchnorm_with_activation(inputs, zero_gamma=False, **bn_act_params, name=name + "preact_")
    else:
        pre_inputs = inputs

    if conv_shortcut:  # Set a new shortcut using conv
        # short_act = activation if attn_block_params["attn_type"] == "bot" else None
        shortcut = conv_shortcut_branch(pre_inputs, filters, preact, strides, shortcut_type, bn_params, name=name)  # activation=None
    else:
        shortcut = layers.MaxPool2D(strides, strides=strides, padding="SAME")(inputs) if strides > 1 else inputs

    deep = deep_branch(pre_inputs, filters, strides, hidden_channel_ratio, use_3x3_kernel, bn_after_attn, bn_act_params, attn_block_params, name=name)

    # print(f">>>> {inputs.shape = }, {shortcut if shortcut is None else shortcut.shape = }, {deep.shape = }, {filters = }, {strides = }")
    if preact:  # ResNetV2
        deep = drop_block(deep, drop_rate)
        return layers.Add(name=name + "output")([shortcut, deep]) if shortcut is not None else deep  # if no shortcut
    else:
        if not (use_3x3_kernel and bn_after_attn):
            deep = batchnorm_with_activation(deep, activation=None, zero_gamma=True, **bn_params, name=name + "3_")
        deep = drop_block(deep, drop_rate)
        out = layers.Add(name=name + "add")([shortcut, deep]) if shortcut is not None else deep  # if no shortcut
        if use_block_output_activation:
            out = activation_by_name(out, activation, name=name + "out_")
        return layers.Activation("linear", name=name + "output")(out)  # Identity, Just need a name here


def aot_stack(
    inputs,
    blocks,
    filters,
    strides=2,
    strides_first=True,
    hidden_channel_ratio=0.25,
    stack_drop=0,
    block_params={},
    attn_types=None,
    attn_params={},
    se_ratio=0,
    use_eca=False,
    groups=1,
    group_size=0,
    name="",
):
    nn = inputs
    # print(">>>> attn_types:", attn_types)
    strides_block_id = 0 if strides_first else blocks - 1
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    for id in range(blocks):
        conv_shortcut = True if id == 0 and (strides != 1 or inputs.shape[channel_axis] != filters) else False
        cur_strides = strides if id == strides_block_id else 1
        block_name = name + "block{}_".format(id + 1)
        block_drop_rate = stack_drop[id] if isinstance(stack_drop, (list, tuple)) else stack_drop
        cur_ratio = hidden_channel_ratio[id] if isinstance(hidden_channel_ratio, (list, tuple)) else hidden_channel_ratio
        attn_block_params = {  # Just save the line width..
            "attn_type": attn_types[id] if isinstance(attn_types, (list, tuple)) else attn_types,
            "attn_params": attn_params[id] if isinstance(attn_params, (list, tuple)) else attn_params,
            "se_ratio": se_ratio[id] if isinstance(se_ratio, (list, tuple)) else se_ratio,
            "use_eca": use_eca[id] if isinstance(use_eca, (list, tuple)) else use_eca,
            "groups": groups[id] if isinstance(groups, (list, tuple)) else groups,
            "group_size": group_size[id] if isinstance(group_size, (list, tuple)) else group_size,
        }
        nn = aot_block(
            nn, filters, cur_strides, conv_shortcut, cur_ratio, block_drop_rate, **block_params, attn_block_params=attn_block_params, name=block_name
        )
    return nn


def deep_stem(inputs, stem_width, use_half_channel=True, activation="relu", last_strides=1, bn_params={}, name=None):
    hidden_channel = stem_width // 2 if use_half_channel else stem_width
    nn = conv2d_no_bias(inputs, hidden_channel, 3, strides=2, padding="same", name=name and name + "1_")
    nn = batchnorm_with_activation(nn, activation=activation, **bn_params, name=name and name + "1_")
    nn = conv2d_no_bias(nn, hidden_channel, 3, strides=1, padding="same", name=name and name + "2_")
    nn = batchnorm_with_activation(nn, activation=activation, **bn_params, name=name and name + "2_")
    nn = conv2d_no_bias(nn, stem_width, 3, strides=last_strides, padding="same", name=name and name + "3_")
    return nn


def quad_stem(inputs, stem_width, activation="relu", stem_act=False, last_strides=2, bn_params={}, name=None):
    nn = conv2d_no_bias(inputs, stem_width // 8, 3, strides=2, padding="same", name=name and name + "1_")
    if stem_act:
        nn = batchnorm_with_activation(nn, activation=activation, **bn_params, name=name and name + "1_")
    nn = conv2d_no_bias(nn, stem_width // 4, 3, strides=1, padding="same", name=name and name + "2_")
    if stem_act:
        nn = batchnorm_with_activation(nn, activation=activation, **bn_params, name=name and name + "2_")
    nn = conv2d_no_bias(nn, stem_width // 2, 3, strides=1, padding="same", name=name and name + "3_")
    nn = batchnorm_with_activation(nn, activation=activation, **bn_params, name=name and name + "3_")
    nn = conv2d_no_bias(nn, stem_width, 3, strides=last_strides, padding="same", name=name and name + "4_")
    return nn


def tiered_stem(inputs, stem_width, activation="relu", last_strides=1, bn_params={}, name=None):
    nn = conv2d_no_bias(inputs, 3 * stem_width // 8, 3, strides=2, padding="same", name=name and name + "1_")
    nn = batchnorm_with_activation(nn, activation=activation, **bn_params, name=name and name + "1_")
    nn = conv2d_no_bias(nn, stem_width // 2, 3, strides=1, padding="same", name=name and name + "2_")
    nn = batchnorm_with_activation(nn, activation=activation, **bn_params, name=name and name + "2_")
    nn = conv2d_no_bias(nn, stem_width, 3, strides=last_strides, padding="same", name=name and name + "3_")
    return nn


def aot_stem(inputs, stem_width, type=None, activation="relu", quad_stem_act=False, last_strides=1, bn_params={}, name=None):
    """stem_type in value `[None, "deep", "deep2", "quad", "tiered"]`."""
    if type == "deep":
        nn = deep_stem(inputs, stem_width, activation=activation, last_strides=last_strides, bn_params=bn_params, name=name)
    elif type == "deep2":  # RegNetZD8_EVO
        nn = deep_stem(inputs, stem_width, use_half_channel=False, activation=activation, last_strides=last_strides, bn_params=bn_params, name=name)
    elif type == "quad":
        nn = quad_stem(inputs, stem_width, activation=activation, stem_act=quad_stem_act, last_strides=last_strides, bn_params=bn_params, name=name)
    elif type == "tiered":
        nn = tiered_stem(inputs, stem_width, activation=activation, last_strides=last_strides, bn_params=bn_params, name=name)
    elif type == "kernel_3x3":
        nn = conv2d_no_bias(inputs, stem_width, 3, strides=2, padding="same", name=name)
    else:
        nn = conv2d_no_bias(inputs, stem_width, 7, strides=2, padding="same", name=name)
    return nn


def AotNet(
    num_blocks,  # Stack parameters
    preact=False,  # False for resnet, True for resnetv2
    strides=[1, 2, 2, 2],
    strides_first=True,  # True for resnet, False for resnetv2
    out_channels=[256, 512, 1024, 2048],
    hidden_channel_ratio=0.25,
    use_3x3_kernel=False,
    use_block_output_activation=True,
    use_evo_norm=False,
    bn_epsilon=1e-5,
    evo_norm_group_size=-1,
    stem_width=64,  # Stem params
    stem_type=None,  # ["deep", "quad", "tiered", "kernel_3x3", None]
    quad_stem_act=False,
    stem_last_strides=1,
    stem_downsample=True,
    attn_types=None,  # Attention block params
    attn_params={},
    se_ratio=0,  # (0, 1)
    use_eca=False,
    groups=1,
    group_size=0,
    bn_after_attn=True,
    shortcut_type="conv",  # shortcut_branch params, ["conv", "avg", "anti_alias", None]
    input_shape=(224, 224, 3),  # Model common params
    num_classes=1000,
    activation="relu",
    drop_connect_rate=0,
    classifier_activation="softmax",
    output_num_features=0,
    dropout=0,
    model_name="aotnet",
    pretrained=None,
    kwargs=None,
):
    """Stem"""
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimention is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(shape=input_shape)
    bn_params = {"use_evo_norm": use_evo_norm, "evo_norm_group_size": evo_norm_group_size, "epsilon": bn_epsilon}
    nn = aot_stem(inputs, stem_width, stem_type, activation, quad_stem_act, last_strides=stem_last_strides, bn_params=bn_params, name="stem_")

    if not preact:
        nn = batchnorm_with_activation(nn, activation=activation, **bn_params, name="stem_")
    if stem_downsample:
        nn = layers.ZeroPadding2D(padding=1, name="stem_pool_pad")(nn)
        nn = layers.MaxPool2D(pool_size=3, strides=2, name="stem_pool")(nn)

    """ Stacks """
    block_params = {  # params same for all blocks
        "preact": preact,
        "use_3x3_kernel": use_3x3_kernel,
        "use_block_output_activation": use_block_output_activation,
        "bn_after_attn": bn_after_attn,
        "shortcut_type": shortcut_type,
        "use_evo_norm": use_evo_norm,
        "evo_norm_group_size": evo_norm_group_size,
        "epsilon": bn_epsilon,
        "activation": activation,
    }

    drop_connect_rates = drop_connect_rates_split(num_blocks, start=0.0, end=drop_connect_rate)
    for id, (num_block, out_channel, stride, drop_connect) in enumerate(zip(num_blocks, out_channels, strides, drop_connect_rates)):
        name = "stack{}_".format(id + 1)
        cur_attn_params = {
            "attn_types": attn_types[id] if isinstance(attn_types, (list, tuple)) else attn_types,
            "attn_params": attn_params[id] if isinstance(attn_params, (list, tuple)) else attn_params,
            "se_ratio": se_ratio[id] if isinstance(se_ratio, (list, tuple)) else se_ratio,
            "use_eca": use_eca[id] if isinstance(use_eca, (list, tuple)) else use_eca,
            "groups": groups[id] if isinstance(groups, (list, tuple)) else groups,
            "group_size": group_size[id] if isinstance(group_size, (list, tuple)) else group_size,
        }
        cur_ratio = hidden_channel_ratio[id] if isinstance(hidden_channel_ratio, (list, tuple)) else hidden_channel_ratio
        nn = aot_stack(nn, num_block, out_channel, stride, strides_first, cur_ratio, drop_connect, block_params, **cur_attn_params, name=name)

    if preact:  # resnetv2 like
        nn = batchnorm_with_activation(nn, activation=activation, zero_gamma=False, **bn_params, name="post_")

    """ Output """
    # nn = output_block(nn, output_num_features, activation, num_classes, drop_rate=dropout, classifier_activation=classifier_activation)
    if output_num_features > 0:
        nn = conv2d_no_bias(nn, output_num_features, 1, strides=1, name="features_")
        nn = batchnorm_with_activation(nn, activation=activation, **bn_params, name="features_")

    if num_classes > 0:
        nn = layers.GlobalAveragePooling2D(name="avg_pool")(nn)
        if dropout > 0:
            nn = layers.Dropout(dropout, name="head_drop")(nn)
        nn = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="predictions")(nn)

    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    # reload_model_weights(model, pretrained_dict={}, sub_release="aotnet", input_shape=input_shape, pretrained=pretrained)
    return model


def AotNet50(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", strides=2, **kwargs):
    num_blocks = [3, 4, 6, 3]
    strides = strides if isinstance(strides, (list, tuple)) else [1, 2, 2, strides]
    return AotNet(**locals(), model_name=kwargs.pop("model_name", "aotnet50"), **kwargs)


def AotNet101(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", strides=2, **kwargs):
    num_blocks = [3, 4, 23, 3]
    strides = strides if isinstance(strides, (list, tuple)) else [1, 2, 2, strides]
    return AotNet(**locals(), model_name=kwargs.pop("model_name", "aotnet101"), **kwargs)


def AotNet152(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", strides=2, **kwargs):
    num_blocks = [3, 8, 36, 3]
    strides = strides if isinstance(strides, (list, tuple)) else [1, 2, 2, strides]
    return AotNet(**locals(), model_name=kwargs.pop("model_name", "aotnet152"), **kwargs)


def AotNet200(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", strides=2, **kwargs):
    num_blocks = [3, 24, 36, 3]
    strides = strides if isinstance(strides, (list, tuple)) else [1, 2, 2, strides]
    return AotNet(**locals(), model_name=kwargs.pop("model_name", "aotnet200"), **kwargs)


def AotNetV2(num_blocks, preact=True, strides=1, strides_first=False, **kwargs):
    strides = strides if isinstance(strides, (list, tuple)) else [2, 2, 2, strides]
    return AotNet(num_blocks, preact=preact, strides=strides, strides_first=strides_first, **kwargs)


def AotNet50V2(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", strides=2, **kwargs):
    num_blocks = [3, 4, 6, 3]
    strides = strides if isinstance(strides, (list, tuple)) else [2, 2, 2, strides]
    return AotNetV2(**locals(), model_name=kwargs.pop("model_name", "aotnet50v2"), **kwargs)


def AotNet101V2(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", strides=2, **kwargs):
    num_blocks = [3, 4, 23, 3]
    strides = strides if isinstance(strides, (list, tuple)) else [2, 2, 2, strides]
    return AotNetV2(**locals(), model_name=kwargs.pop("model_name", "aotnet101v2"), **kwargs)


def AotNet152V2(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", strides=2, **kwargs):
    num_blocks = [3, 8, 36, 3]
    strides = strides if isinstance(strides, (list, tuple)) else [2, 2, 2, strides]
    return AotNetV2(**locals(), model_name=kwargs.pop("model_name", "aotnet152v2"), **kwargs)


def AotNet200V2(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", strides=2, **kwargs):
    num_blocks = [3, 24, 36, 3]
    strides = strides if isinstance(strides, (list, tuple)) else [2, 2, 2, strides]
    return AotNetV2(**locals(), model_name=kwargs.pop("model_name", "aotnet200v2"), **kwargs)



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format, initializers
from keras_cv_attention_models.download_and_load import reload_model_weights
from keras_cv_attention_models.attention_layers import ClassToken, layer_norm, conv2d_no_bias, activation_by_name, add_pre_post_process


PRETRAINED_DICT = {
    "coat_lite_tiny": {"imagenet": "e45487e7bfb44faac97b1af51f8bbd01"},
    "coat_lite_mini": {"imagenet": "e5e3f5e4b86765ee75f8bf03973d70a0"},
    "coat_lite_small": {"imagenet": "eddffc46a64eb0a21b7ecc057f231756"},
    "coat_tiny": {"imagenet": "6418d9580ad9ea0a6755c77d8d7bad49"},
    "coat_mini": {"imagenet": "dc284967f6bd32df8e1e03074b2d773d"},
}


# Not a layer, just for reusable
class ConvPositionalEncoding:
    def __init__(self, kernel_size=3, input_height=-1, name=None):
        self.kernel_size, self.input_height, self.name = kernel_size, input_height, name
        if image_data_format() == "channels_last":
            self.pad = [[0, 0], [kernel_size // 2, kernel_size // 2], [kernel_size // 2, kernel_size // 2], [0, 0]]
        else:
            self.pad = [[0, 0], [0, 0], [kernel_size // 2, kernel_size // 2], [kernel_size // 2, kernel_size // 2]]
        self.built = False

    def build(self, input_shape):
        self.height = self.input_height if self.input_height > 0 else int(float(input_shape[1] - 1) ** 0.5)
        self.width = (input_shape[1] - 1) // self.height

        self.channel = input_shape[-1]
        self.dconv = layers.DepthwiseConv2D(self.kernel_size, strides=1, padding="VALID", name=self.name and self.name + "depth_conv")

    def __call__(self, inputs, **kwargs):
        if not self.built:
            self.build(inputs.shape)
            self.built = True

        cls_token, img_token = inputs[:, :1], inputs[:, 1:]
        img_token = functional.reshape(img_token, [-1, self.height, self.width, self.channel])
        nn = img_token if image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(img_token)
        # print(f"{nn.shape = }")
        nn = self.dconv(functional.pad(nn, self.pad))
        nn = nn if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(nn)
        nn = layers.Add()([nn, img_token])
        nn = functional.reshape(nn, [-1, self.height * self.width, self.channel])
        return functional.concat([cls_token, nn], axis=1)


# Not a layer, just for reusable
class ConvRelativePositionalEncoding:
    def __init__(self, head_splits=[2, 3, 3], head_kernel_size=[3, 5, 7], input_height=-1, name=None):
        self.head_splits, self.head_kernel_size, self.input_height, self.name = head_splits, head_kernel_size, input_height, name
        self.built = False

    def build(self, query_shape):
        # print(query_shape)
        self.height = self.input_height if self.input_height > 0 else int(float(query_shape[2] - 1) ** 0.5)
        self.width = (query_shape[2] - 1) // self.height
        self.num_heads, self.query_dim = query_shape[1], query_shape[-1]
        self.channel_splits = [ii * self.query_dim for ii in self.head_splits]

        self.dconvs = []
        self.pads = []
        for id, (head_split, kernel_size) in enumerate(zip(self.head_splits, self.head_kernel_size)):
            name = self.name and self.name + "depth_conv_" + str(id + 1)
            dconv = layers.DepthwiseConv2D(kernel_size, strides=1, padding="VALID", name=name)
            if image_data_format() == "channels_last":
                pad = [[0, 0], [kernel_size // 2, kernel_size // 2], [kernel_size // 2, kernel_size // 2], [0, 0]]
            else:
                pad = [[0, 0], [0, 0], [kernel_size // 2, kernel_size // 2], [kernel_size // 2, kernel_size // 2]]
            self.dconvs.append(dconv)
            self.pads.append(pad)

    def __call__(self, query, value, **kwargs):
        if not self.built:
            self.build(query.shape)
            self.built = True
        img_token_q, img_token_v = query[:, :, 1:, :], value[:, :, 1:, :]

        if image_data_format() == "channels_last":
            img_token_v = functional.transpose(img_token_v, [0, 2, 1, 3])  # [batch, blocks, num_heads, query_dim]
            img_token_v = functional.reshape(img_token_v, [-1, self.height, self.width, self.num_heads * self.query_dim])
            split_values = functional.split(img_token_v, self.channel_splits, axis=-1)
        else:
            img_token_v = functional.transpose(img_token_v, [0, 1, 3, 2])  # [batch, num_heads, query_dim, blocks]
            img_token_v = functional.reshape(img_token_v, [-1, self.num_heads * self.query_dim, self.height, self.width])
            split_values = functional.split(img_token_v, self.channel_splits, axis=1)
        nn = [dconv(functional.pad(split_value, pad)) for split_value, dconv, pad in zip(split_values, self.dconvs, self.pads)]

        if image_data_format() == "channels_last":
            nn = functional.concat(nn, axis=-1)
            conv_v_img = functional.reshape(nn, [-1, self.height * self.width, self.num_heads, self.query_dim])
            conv_v_img = functional.transpose(conv_v_img, [0, 2, 1, 3])
        else:
            nn = functional.concat(nn, axis=1)
            conv_v_img = functional.reshape(nn, [-1, self.num_heads, self.query_dim, self.height * self.width])
            conv_v_img = functional.transpose(conv_v_img, [0, 1, 3, 2])

        EV_hat_img = img_token_q * conv_v_img
        return functional.pad(EV_hat_img, [[0, 0], [0, 0], [1, 0], [0, 0]])


def factor_attention_conv_relative_positional_encoding(inputs, shared_crpe=None, num_heads=8, qkv_bias=True, name=""):
    blocks, dim = inputs.shape[1], inputs.shape[-1]
    key_dim = dim // num_heads
    qk_scale = 1.0 / (float(key_dim) ** 0.5)

    qkv = layers.Dense(dim * 3, use_bias=qkv_bias, name=name + "qkv")(inputs)
    qkv = layers.Reshape([blocks, 3, num_heads, key_dim])(qkv)
    qq, kk, vv = functional.transpose(qkv, [2, 0, 3, 1, 4])  # [qkv, batch, num_heads, blocks, key_dim]
    # print(f">>>> {qkv.shape = }, {qq.shape = }, {kk.shape = }, {vv.shape = }")

    # Factorized attention.
    # kk = tf.nn.softmax(kk, axis=2)  # On `blocks` dimension
    kk = layers.Softmax(axis=2, name=name and name + "attention_scores")(kk)  # On `blocks` dimension
    kk = functional.transpose(kk, [0, 1, 3, 2])
    factor_att = qq @ (kk @ vv)

    # Convolutional relative position encoding.
    crpe_out = shared_crpe(qq, vv) if shared_crpe is not None else ConvRelativePositionalEncoding(name=name + "crpe_")(qq, vv)

    # Merge and reshape.
    nn = layers.Add()([factor_att * qk_scale, crpe_out])
    nn = layers.Permute([2, 1, 3])(nn)
    nn = layers.Reshape([blocks, dim])(nn)
    nn = layers.Dense(dim, name=name + "out")(nn)
    return nn


def cpe_norm_crpe(inputs, shared_cpe=None, shared_crpe=None, num_heads=8, name=""):
    cpe_out = shared_cpe(inputs) if shared_cpe is not None else ConvPositionalEncoding(name=name + "cpe_")(inputs)  # shared
    nn = layer_norm(cpe_out, axis=-1, name=name + "norm1")
    crpe_out = factor_attention_conv_relative_positional_encoding(nn, shared_crpe=shared_crpe, num_heads=num_heads, name=name + "factoratt_crpe_")
    return cpe_out, crpe_out


def res_mlp_block(cpe_out, crpe_out, mlp_ratio=4, drop_rate=0, activation="gelu", name=""):
    if drop_rate > 0:
        crpe_out = layers.Dropout(drop_rate, noise_shape=(None, 1, 1), name=name + "drop_1")(crpe_out)
    cpe_crpe = layers.Add()([cpe_out, crpe_out])

    # MLP
    pre_mlp = layer_norm(cpe_crpe, axis=-1, name=name + "norm2")
    nn = layers.Dense(pre_mlp.shape[-1] * mlp_ratio, name=name + "mlp_dense_0")(pre_mlp)
    nn = activation_by_name(nn, activation, name=name + "mlp_")
    nn = layers.Dense(pre_mlp.shape[-1], name=name + "mlp_dense_1")(nn)

    if drop_rate > 0:
        nn = layers.Dropout(drop_rate, noise_shape=(None, 1, 1), name=name + "drop_2")(nn)
    return layers.Add(name=name + "output")([cpe_crpe, nn])


def serial_block(inputs, embed_dim, shared_cpe=None, shared_crpe=None, num_heads=8, mlp_ratio=4, drop_rate=0, activation="gelu", name=""):
    cpe_out, crpe_out = cpe_norm_crpe(inputs, shared_cpe, shared_crpe, num_heads, name=name)
    out = res_mlp_block(cpe_out, crpe_out, mlp_ratio, drop_rate, activation=activation, name=name)
    return out


def resample(image, target_shape, class_token=None):
    # print(f"{image.shape = }, {target_shape = }")
    image = image if image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(image)
    # out_image = functional.cast(functional.resize(image, target_shape, method="bilinear"), image.dtype)
    out_image = functional.resize(image, target_shape, method="bilinear")
    out_image = out_image if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(out_image)

    if class_token is not None:
        out_image = functional.reshape(out_image, [-1, out_image.shape[1] * out_image.shape[2], out_image.shape[-1]])
        return functional.concat([class_token, out_image], axis=1)
    else:
        return out_image


def parallel_block(inputs, shared_cpes=None, shared_crpes=None, block_heights=[], num_heads=8, mlp_ratios=[], drop_rate=0, activation="gelu", name=""):
    # Conv-Attention.
    # print(f">>>> {block_heights = }")
    cpe_outs, crpe_outs, crpe_images, resample_shapes = [], [], [], []
    block_heights = block_heights[1:]
    for id, (xx, shared_cpe, shared_crpe) in enumerate(zip(inputs[1:], shared_cpes[1:], shared_crpes[1:])):
        cur_name = name + "{}_".format(id + 2)
        cpe_out, crpe_out = cpe_norm_crpe(xx, shared_cpe, shared_crpe, num_heads, name=cur_name)
        cpe_outs.append(cpe_out)
        crpe_outs.append(crpe_out)
        height = block_heights[id] if len(block_heights) > id else int(float(crpe_out.shape[1] - 1) ** 0.5)
        width = (crpe_out.shape[1] - 1) // height
        crpe_images.append(functional.reshape(crpe_out[:, 1:, :], [-1, height, width, crpe_out.shape[-1]]))
        resample_shapes.append([height, width])
        # print(f">>>> {crpe_out.shape = }, {crpe_images[-1].shape = }")
    crpe_stack = [  # [[None, 28, 28, 152], [None, 14, 14, 152], [None, 7, 7, 152]]
        crpe_outs[0] + resample(crpe_images[1], resample_shapes[0], crpe_outs[1][:, :1]) + resample(crpe_images[2], resample_shapes[0], crpe_outs[2][:, :1]),
        crpe_outs[1] + resample(crpe_images[2], resample_shapes[1], crpe_outs[2][:, :1]) + resample(crpe_images[0], resample_shapes[1], crpe_outs[0][:, :1]),
        crpe_outs[2] + resample(crpe_images[1], resample_shapes[2], crpe_outs[1][:, :1]) + resample(crpe_images[0], resample_shapes[2], crpe_outs[0][:, :1]),
    ]

    # MLP
    outs = []
    for id, (cpe_out, crpe_out, mlp_ratio) in enumerate(zip(cpe_outs, crpe_stack, mlp_ratios[1:])):
        cur_name = name + "{}_".format(id + 2)
        out = res_mlp_block(cpe_out, crpe_out, mlp_ratio, drop_rate, activation=activation, name=cur_name)
        outs.append(out)
    return inputs[:1] + outs  # inputs[0] directly out


def patch_embed(inputs, embed_dim, patch_size=2, input_height=-1, name=""):
    if len(inputs.shape) == 3:
        input_height = input_height if input_height > 0 else int(float(inputs.shape[1]) ** 0.5)
        input_width = inputs.shape[1] // input_height
        inputs = layers.Reshape([input_height, input_width, inputs.shape[-1]])(inputs)
        inputs = inputs if image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(inputs)
    nn = conv2d_no_bias(inputs, embed_dim, kernel_size=patch_size, strides=patch_size, use_bias=True, name=name)
    nn = nn if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(nn)
    block_height = nn.shape[1]
    nn = layers.Reshape([nn.shape[1] * nn.shape[2], nn.shape[3]])(nn)  # flatten(2)
    nn = layer_norm(nn, axis=-1, name=name)
    return nn, block_height


def CoaT(
    serial_depths=[2, 2, 2, 2],
    embed_dims=[64, 128, 256, 320],
    mlp_ratios=[8, 8, 4, 4],
    parallel_depth=0,
    patch_size=4,
    num_heads=8,
    head_splits=[2, 3, 3],
    head_kernel_size=[3, 5, 7],
    use_shared_cpe=True,  # For checking model architecture only, keep input_shape height == width if set False
    use_shared_crpe=True,  # For checking model architecture only, keep input_shape height == width if set False
    out_features=None,
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="gelu",
    drop_connect_rate=0,
    classifier_activation="softmax",
    pretrained="imagenet",
    model_name="coat",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimention is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)

    # serial blocks
    nn = inputs
    classfier_outs = []
    shared_cpes = []
    shared_crpes = []
    block_heights = []
    for sid, (depth, embed_dim, mlp_ratio) in enumerate(zip(serial_depths, embed_dims, mlp_ratios)):
        name = "serial{}_".format(sid + 1)
        patch_size = patch_size if sid == 0 else 2
        patch_input_height = -1 if sid == 0 else block_heights[-1]
        # print(f">>>> {nn.shape = }")
        nn, block_height = patch_embed(nn, embed_dim, patch_size=patch_size, input_height=patch_input_height, name=name + "patch_")
        block_heights.append(block_height)
        # print(f">>>> {nn.shape = }, {block_height = }")
        nn = ClassToken(name=name + "class_token")(nn)
        shared_cpe = ConvPositionalEncoding(kernel_size=3, input_height=block_height, name="cpe{}_".format(sid + 1)) if use_shared_cpe else None
        shared_crpe = ConvRelativePositionalEncoding(head_splits, head_kernel_size, block_height, name="crpe{}_".format(sid + 1)) if use_shared_crpe else None
        for bid in range(depth):
            block_name = name + "block{}_".format(bid + 1)
            nn = serial_block(nn, embed_dim, shared_cpe, shared_crpe, num_heads, mlp_ratio, activation=activation, name=block_name)
        classfier_outs.append(nn)
        shared_cpes.append(shared_cpe)
        shared_crpes.append(shared_crpe)
        nn = nn[:, 1:, :]  # remove class token

    # Parallel blocks.
    for pid in range(parallel_depth):
        name = "parallel{}_".format(pid + 1)
        classfier_outs = parallel_block(classfier_outs, shared_cpes, shared_crpes, block_heights, num_heads, mlp_ratios, activation=activation, name=name)

    if out_features is not None:  # Return intermediate features (for down-stream tasks).
        nn = [classfier_outs[id][:, 1:, :] for id in out_features]
    elif parallel_depth == 0:  # Lite model, only serial blocks, Early return.
        nn = layer_norm(classfier_outs[-1], axis=-1, name="out_")[:, 0]
    else:
        nn = [layer_norm(xx, axis=-1, name="out_{}_".format(id + 1))[:, :1, :] for id, xx in enumerate(classfier_outs[1:])]
        nn = layers.Concatenate(axis=1)(nn)
        nn = layers.Permute([2, 1])(nn) if image_data_format() == "channels_last" else nn
        nn = layers.Conv1D(1, 1, name="aggregate")(nn)
        nn = nn[:, :, 0] if image_data_format() == "channels_last" else nn[:, 0]
    if out_features is None and num_classes > 0:
        nn = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="predictions")(nn)

    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, pretrained_dict=PRETRAINED_DICT, sub_release="coat", pretrained=pretrained)
    return model


def CoaTLiteTiny(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return CoaT(**locals(), model_name="coat_lite_tiny", **kwargs)


def CoaTLiteMini(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dims = [64, 128, 320, 512]
    return CoaT(**locals(), model_name="coat_lite_mini", **kwargs)


def CoaTLiteSmall(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    serial_depths = [3, 4, 6, 3]
    embed_dims = [64, 128, 320, 512]
    return CoaT(**locals(), model_name="coat_lite_small", **kwargs)


def CoaTTiny(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dims = [152, 152, 152, 152]
    mlp_ratios = [4, 4, 4, 4]
    parallel_depth = 6
    return CoaT(**locals(), model_name="coat_tiny", **kwargs)


def CoaTMini(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dims = [152, 216, 216, 216]
    mlp_ratios = [4, 4, 4, 4]
    parallel_depth = 6
    return CoaT(**locals(), model_name="coat_mini", **kwargs)



from keras_cv_attention_models.botnet.botnet import BotNet, BotNet50, BotNet101, BotNet152, BotNet26T, BotNextECA26T, BotNetSE33T, RelativePositionalEmbedding, mhsa_with_relative_position_embedding


__head_doc__ = """
Keras implementation of [botnet](https://gist.github.com/aravindsrinivas/56359b79f0ce4449bcb04ab4b56a57a2).
Paper [PDF 2101.11605 Bottleneck Transformers for Visual Recognition](https://arxiv.org/pdf/2101.11605.pdf).
"""

__tail_doc__ = """  strides: strides used in the last stack. It's reported `1` works better than `2`, but slower.
  input_shape: it should have exactly 3 inputs channels, default `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `relu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `softmax`.
  pretrained: one of `None` (random initialization) or 'imagenet' (pre-training on ImageNet).
      Will try to download and load pre-trained model weights if not None.
  **kwargs: other parameters from `AotNet` if not conflict.

Returns:
    A `keras.Model` instance.
"""

BotNet.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model         | Params | FLOPs  | Input | Top1 Acc |
  | ------------- | ------ | ------ | ----- | -------- |
  | BotNet50      | 21M    | 5.42G  | 224   |          |
  | BotNet101     | 41M    | 9.13G  | 224   |          |
  | BotNet152     | 56M    | 12.84G | 224   |          |
  | BotNet26T     | 12.5M  | 3.30G  | 256   | 79.246   |
  | BotNextECA26T | 10.59M | 2.45G  | 256   | 79.270   |
  | BotNetSE33T   | 13.7M  | 3.89G  | 256   | 81.2     |
"""

BotNet50.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

BotNet101.__doc__ = BotNet50.__doc__
BotNet152.__doc__ = BotNet50.__doc__

BotNet26T.__doc__ = __head_doc__ + """Model weights are reloaded from timm [Github rwightman/pytorch-image-models](https://github.com/rwightman/pytorch-image-models).

Args:
""" + __tail_doc__
BotNextECA26T.__doc__ = BotNet26T.__doc__
BotNetSE33T.__doc__ = BotNet26T.__doc__

mhsa_with_relative_position_embedding.__doc__ = __head_doc__ + """
Multi head self attention with positional embedding. Defined as function, not layer.
Typical MHSA with `RelativePositionalEmbedding` added to `attention_scores`.

Args:
  inputs: input tensor.
  num_heads: Number of attention heads.
  key_dim: Size of each attention head for query and key. Default `0` for `key_dim = inputs.shape[-1] // num_heads`.
  relative: Boolean, whether using relative or absolute positional embedding.
  out_weight: Boolean, whether use an ouput dense.
  out_bias: Boolean, whether the ouput dense layer use bias vectors/matrices.
  out_shape: The expected shape of an output tensor. If not specified, projects back to the input dim.
  attn_dropout: Dropout probability for attention.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([14, 16, 256])
>>> nn = attention_layers.mhsa_with_relative_position_embedding(inputs, num_heads=4)
>>> print(f"{nn.shape = }")
nn.shape = TensorShape([None, 14, 16, 256])

>>> mm = keras.models.Model(inputs, nn)
>>> mm.summary()
>>> print({ii.name: ii.shape for ii in mm.weights})
{'dense/kernel:0': TensorShape([1, 1, 256, 768]),
 'relative_positional_embedding/r_height:0': TensorShape([64, 27]),
 'relative_positional_embedding/r_width:0': TensorShape([64, 31]),
 'dense_1/kernel:0': TensorShape([256, 256])}
"""

RelativePositionalEmbedding.__doc__ = __head_doc__ + """
Relative Positional Embedding layer. Supports also absolute positional embedding.
Supports both `absolute` / `relative` positional embedding.
Layer weights is dotted with input generating positional embedding.
It's using same value for all headers.

input: `[batch, num_heads, height, width, key_dim]`.
output: `[batch, num_heads, height, width, position_height, position_width]`

Args:
  position_height: positional embedding height. Default `0` for using `input_shape[2]`.
      Should be larger than `input_shape[2]`.
  position_width: positional embedding width. Default `0` for using `input_shape[3]` or `position_height` if set.
      Should be larger than `input_shape[3]`.
  use_absolute_pos: Set `True` to use absolute positional embeddings.
  dynamic_shape: Set `True` for dynamically change output shape depending on inputs shape.
      - Works only if coming inputs shape is smaller than orignal initialized `position_height` and `position_width`.
      - For larger inputs, please reload layer weights by `self.load_resized_weights`.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> aa = attention_layers.RelativePositionalEmbedding()
>>> print(f"{aa(tf.ones([1, 4, 14, 16, 32])).shape = }")
aa(tf.ones([1, 4, 14, 16, 32])).shape = TensorShape([1, 4, 14, 16, 14, 16])
>>> print(f"{aa(tf.ones([1, 4, 4, 6, 32])).shape = }")  # last 2 dimension in output is `[position_height, position_width]`
aa(tf.ones([1, 4, 4, 6, 32])).shape = TensorShape([1, 4, 4, 6, 14, 16])

>>> print({ii.name:ii.shape for ii in aa.weights})
{'relative_positional_embedding_6/r_height:0': TensorShape([32, 27]),
 'relative_positional_embedding_6/r_width:0': TensorShape([32, 31])}

For `dynamic_shape=True`:
>>> aa = attention_layers.RelativePositionalEmbedding(dynamic_shape=True)
>>> print(f"{aa(tf.ones([1, 4, 14, 16, 32])).shape = }")
aa(tf.ones([1, 4, 14, 16, 32])).shape = TensorShape([1, 4, 14, 16, 14, 16])
>>> print(f"{aa(tf.ones([1, 4, 4, 6, 32])).shape = }")  # last 2 dimension in output is `[height, width]`
aa(tf.ones([1, 4, 4, 6, 32])).shape = TensorShape([1, 4, 4, 6, 4, 6])

Reload layer weights by `self.load_resized_weights`:
>>> bb = attention_layers.RelativePositionalEmbedding(dynamic_shape=True)
>>> bb.build([None, 4, 24, 26, 32])
>>> print({ii.name:ii.shape for ii in bb.weights})
{'r_height:0': TensorShape([32, 47]),
 'r_width:0': TensorShape([32, 51])}
>>> bb.load_resized_weights(aa)
>>> print(f"{bb(tf.ones([1, 4, 21, 22, 32])).shape = }")
bb(tf.ones([1, 4, 21, 22, 32])).shape = TensorShape([1, 4, 21, 22, 21, 22])
"""



import tensorflow as tf
from tensorflow import keras


class TokenLabelAlign:
    """
    >>> dd = TokenLabelAlign(target_num_pathes=7)
    >>> print(f"{np.allclose(tf.math.top_k(dd(cc[0]), 5)[0], cc[0][1]) = }")
    >>> print(f"{np.allclose(tf.math.top_k(dd(cc[0]), 5)[1], cc[0][0]) = }")
    """

    def __init__(self, num_classes=10, target_num_pathes=14, align_method="bilinear"):
        self.num_classes, self.align_method = num_classes, align_method
        target_num_pathes = target_num_pathes[:2] if isinstance(target_num_pathes, (list, tuple)) else (target_num_pathes, target_num_pathes)
        self.target_num_pathes_h, self.target_num_pathes_w = target_num_pathes
        self.built = False

    def build(self, token_label_shape):
        # To one-hot
        self.source_num_pathes_h, self.source_num_pathes_w, num_topk = token_label_shape[1], token_label_shape[2], token_label_shape[3]
        self.target_num_pathes_h = self.target_num_pathes_h if self.target_num_pathes_h > 0 else self.source_num_pathes_h
        self.target_num_pathes_w = self.target_num_pathes_w if self.target_num_pathes_w > 0 else self.source_num_pathes_w
        hh, ww = tf.meshgrid(range(self.source_num_pathes_h), range(self.source_num_pathes_w), indexing="ij")
        hhww = tf.concat([tf.reshape(hh, [-1, 1, 1]), tf.reshape(ww, [-1, 1, 1])], axis=-1)
        self.one_hot_hhww = tf.repeat(hhww, num_topk, axis=1)

        # Align to target shape
        hh, ww = tf.meshgrid(range(0, self.target_num_pathes_h), range(0, self.target_num_pathes_w), indexing="ij")
        hh, ww = tf.reshape(hh, [-1, 1]), tf.reshape(ww, [-1, 1])
        boxes = tf.concat([hh, ww, hh + 1, ww + 1], axis=-1)
        self.boxes = tf.cast(boxes, "float32") / [self.target_num_pathes_h, self.target_num_pathes_w, self.target_num_pathes_h, self.target_num_pathes_w]
        self.box_indices = [0] * (self.target_num_pathes_h * self.target_num_pathes_w)  # 0 is indicating which batch
        # self.need_align = self.target_num_pathes_h != self.source_num_pathes_h or self.target_num_pathes_w != self.source_num_pathes_w
        self.built = True

        source_num_pathes, target_num_pathes = [self.source_num_pathes_h, self.source_num_pathes_w], [self.target_num_pathes_h, self.target_num_pathes_w]
        print(">>>> [TokenLabelAlign] source_num_pathes: {}, target_num_pathes: {}".format(source_num_pathes, target_num_pathes))

    def __call__(self, token_label, flip_left_right=False, scale_hh=1, scale_ww=1, crop_hh=0, crop_ww=0):
        if not self.built:
            self.build(token_label.shape)
        label_pos, label_score = tf.cast(token_label[0], "int32"), tf.cast(token_label[1], "float32")
        label_position = tf.concat([tf.reshape(self.one_hot_hhww, [-1, 2]), tf.reshape(label_pos, [-1, 1])], axis=-1)
        token_label_one_hot = tf.zeros([self.source_num_pathes_h, self.source_num_pathes_w, self.num_classes])
        token_label_one_hot = tf.tensor_scatter_nd_update(token_label_one_hot, label_position, tf.reshape(label_score, [-1]))

        token_label_one_hot = tf.cond(flip_left_right, lambda: tf.image.flip_left_right(token_label_one_hot), lambda: token_label_one_hot)
        boxes = (self.boxes + [crop_hh, crop_ww, crop_hh, crop_ww]) / [scale_hh, scale_ww, scale_hh, scale_ww]
        # if self.need_align:
        token_label_one_hot = tf.expand_dims(token_label_one_hot, 0)  # Expand a batch dimension, required by crop_and_resize
        token_label_one_hot = tf.image.crop_and_resize(token_label_one_hot, boxes, self.box_indices, crop_size=(1, 1), method=self.align_method)
        return tf.reshape(token_label_one_hot, (self.target_num_pathes_h, self.target_num_pathes_w, self.num_classes))


def build_token_label_file(
    data_name, model, input_shape=-1, batch_size=16, rescale_mode="auto", resize_method="bicubic", resize_antialias=True, token_label_top_k=5, save_path=None
):
    import pickle
    import numpy as np
    from tqdm import tqdm
    from keras_cv_attention_models.imagenet import data, train_func
    from keras_cv_attention_models import model_surgery

    total_images, num_classes, steps_per_epoch, num_channels = data.init_dataset(data_name, batch_size=batch_size, info_only=True)
    if isinstance(model, str) and model.endswith(".h5"):
        model = keras.models.load_model(model, compile=False)
    else:
        model = train_func.init_model(model, input_shape, num_classes)
    token_label_model = model_surgery.convert_to_token_label_model(model)

    if isinstance(rescale_mode, str) and rescale_mode.lower() == "auto":
        rescale_mode = getattr(model, "rescale_mode", "torch")
        print(">>>> rescale_mode:", rescale_mode)

    train_dataset = data.init_dataset(
        data_name,
        input_shape=model.input_shape[1:],
        batch_size=batch_size,
        rescale_mode=rescale_mode,
        resize_method=resize_method,
        resize_antialias=resize_antialias,
        magnitude=-1,  # Disable random_flip_left_right
        use_shuffle=False,  # Disable shuffle
    )[0]
    if save_path is None:
        data_name = data_name.replace("/", "_")
        patches = token_label_model.output_shape[1][1:-1]
        save_path = "{}_{}_{}_{}_{}.pkl".format(data_name, model.name, patches[0], patches[1], token_label_top_k)

    rrs = []
    need_softmax = False if token_label_model.layers[-1].activation.__name__ == "softmax" else True
    for image_batch, _ in tqdm(train_dataset):
        predictions = token_label_model(image_batch)[-1]
        if need_softmax:
            predictions = tf.nn.softmax(predictions, axis=-1)
        prediction_scores, prediction_ids = tf.math.top_k(predictions, k=token_label_top_k)
        rr = tf.stack([tf.cast(prediction_ids, prediction_scores.dtype), prediction_scores], axis=1)
        rrs.append(rr.numpy())
    rrs = np.concatenate(rrs, axis=0)
    # np.save(label_token_file, rrs)
    with open(save_path, "wb") as ff:
        pickle.dump(rrs, ff)
    print(">>>> Saved to:", save_path)



from keras_cv_attention_models.tinyvit.tinyvit import TinyViT, TinyViT_5M, TinyViT_11M, TinyViT_21M

__head_doc__ = """
Keras implementation of [Github microsoft/TinyViT](https://github.com/microsoft/Cream/tree/main/TinyViT).
Paper [PDF 2207.10666 TinyViT: Fast Pretraining Distillation for Small Vision Transformers](https://arxiv.org/pdf/2207.10666.pdf).
"""

__tail_doc__ = """  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `gelu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be a constant value like `0.2`,
      or a tuple value like `(0, 0.2)` indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  layer_scale: int value indicates layer scale init value for each stack. Default 0 for not using.
      [Going deeper with Image Transformers](https://arxiv.org/abs/2103.17239).
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `None`.
  pretrained: one of `None` (random initialization) or 'imagenet' or 'imagenet21k-ft1k' (pre-training on ImageNet21k and fine-tuned ImageNet).
      Will try to download and load pre-trained model weights if not None.

Returns:
    A `keras.Model` instance.
"""

TinyViT.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  out_channels: output channels for each stack.
  block_types: block types for each stack,
      - `conv` or any `c` / `C` starts word, means `mlp_block_with_depthwise_conv` block.
      - `transfrom` or any `t` / `T` starts word, means `multi_head_self_attention` block.
      value could be in format like `"cctt"` or `"CCTT"` or `["conv", "conv", "transfrom", "transform"]`.
  num_heads: int or list of int value indicates attention heads number for each transformer stack.
  window_ratios: int or list of int value indicates attention heads window ratio number for each transformer stack.
      Actual using `window_size = ceil(cur_input_shape / window_ratio)`.
      For `input_shape=(224, 224, 3)` will be window_sizes=[7, 7, 14, 7], for `(384, 384, 3)` will be `[12, 12, 24, 12]`.
  mlp_ratio: int value indicates expand ratio for mlp blocks hidden channel in each stack.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model                | Params | FLOPs | Input | Top1 Acc |
  | -------------------- | ------ | ----- | ----- | -------- |
  | TinyViT_5M, distill  | 5.4M   | 1.3G  | 224   | 79.1     |
  | - imagenet21k-ft1k   | 5.4M   | 1.3G  | 224   | 80.7     |
  | TinyViT_11M, distill | 11M    | 2.0G  | 224   | 81.5     |
  | - imagenet21k-ft1k   | 11M    | 2.0G  | 224   | 83.2     |
  | TinyViT_21M, distill | 21M    | 4.3G  | 224   | 83.1     |
  | - imagenet21k-ft1k   | 21M    | 4.3G  | 224   | 84.8     |
  |                      | 21M    | 13.8G | 384   | 86.2     |
  |                      | 21M    | 27.0G | 512   | 86.5     |
"""

TinyViT_5M.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

TinyViT_11M.__doc__ = TinyViT_5M.__doc__
TinyViT_21M.__doc__ = TinyViT_5M.__doc__



""" Setup
"""
from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

exec(open("keras_cv_attention_models/version.py").read())
setup(
    name="keras-cv-attention-models",
    version=__version__,
    description="Tensorflow keras computer vision attention models. Alias kecam. https://github.com/leondgarse/keras_cv_attention_models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/leondgarse/keras_cv_attention_models",
    author="Leondgarse",
    author_email="leondgarse@google.com",
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    # Note that this is a string of words separated by whitespace, not a list.
    keywords="tensorflow keras cv attention pretrained models kecam",
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    install_requires=[
        # "tensorflow-macos;platform_system=='Darwin'",  # [???]
        "tensorflow;platform_system!='Darwin'",
        "tensorflow-addons;platform_machine!='aarch64' and platform_machine!='aarch32'",
        "tensorflow-datasets;platform_machine!='aarch64' and platform_machine!='aarch32'"  # >4.7.0 needs dm-tree, failed on arm, just skip
    ],
    python_requires=">=3.6",
    license="Apache 2.0",
)



from keras_cv_attention_models.beit.beit import Beit, keras_model_load_weights_from_pytorch_model


def ViT(
    depth=12,
    embed_dim=768,
    num_heads=12,
    mlp_ratio=4,
    patch_size=16,
    attn_key_dim=0,
    attn_qv_bias=False,  # Deault False for Vit, True for Beit, if True and attn_qkv_bias seing False, will add BiasLayer for query and key.
    attn_qkv_bias=True,  # True for Vit, False for Beit, if True, will just use bias in qkv_dense, and set qv_bias False.
    attn_out_weight=True,
    attn_out_bias=True,
    attn_dropout=0,
    gamma_init_value=0,  # 0 for Vit, 0.1 for Beit, if > 0 will use `layer_scale` on block output
    use_abs_pos_emb=True,  # True for Vit, False for Beit, whether use abcolute positional embedding or relative one in attention blocks
    use_abs_pos_emb_on_cls_token=True,  # False for FlexiViT, no_embed_class in timm. If use_abs_pos_emb is True, whether apply pos_emb on cls_token.
    use_mean_pooling=False,  # False for Vit, True for Beit, whether use use mean output or `class_token` output
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="gelu",
    drop_connect_rate=0,
    classifier_activation="softmax",
    pretrained=None,
    force_reload_mismatch=False,  # set True if patch_size changed, will force reloading pos_emb and stem_conv weights
    model_name="vit",
    kwargs=None,
):
    return Beit(**locals())


""" FlexiViT """


def FlexiViTSmall(input_shape=(240, 240, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dim = 384
    depth = 12
    num_heads = 6
    use_abs_pos_emb_on_cls_token = False  # no_embed_class in timm
    force_reload_mismatch = kwargs.get("patch_size", 16) != 16  # If patch_size not 16, force reload pos_emb and stem_conv weights
    return ViT(**locals(), model_name="flexivit_small", **kwargs)


def FlexiViTBase(input_shape=(240, 240, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dim = 768
    depth = 12
    num_heads = 12
    use_abs_pos_emb_on_cls_token = False  # no_embed_class in timm
    force_reload_mismatch = kwargs.get("patch_size", 16) != 16  # If patch_size not 16, force reload pos_emb and stem_conv weights
    return ViT(**locals(), model_name="flexivit_base", **kwargs)


def FlexiViTLarge(input_shape=(240, 240, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dim = 1024
    depth = 24
    num_heads = 16
    use_abs_pos_emb_on_cls_token = False  # no_embed_class in timm
    force_reload_mismatch = kwargs.get("patch_size", 16) != 16  # If patch_size not 16, force reload pos_emb and stem_conv weights
    return Beit(**locals(), model_name="flexivit_large", **kwargs)



from keras_cv_attention_models.yolov7.yolov7 import (
    YOLOV7Backbone,
    YOLOV7,
    YOLOV7_Tiny,
    YOLOV7_CSP,
    YOLOV7_X,
    YOLOV7_W6,
    YOLOV7_E6,
    YOLOV7_D6,
    YOLOV7_E6E,
)

__head_doc__ = """
Keras implementation of [Github WongKinYiu/yolov7](https://github.com/WongKinYiu/yolov7).
Paper [Paper 2207.02696 YOLOv7: Trainable bag-of-freebies sets new state-of-the-art for real-time object detectors](https://arxiv.org/pdf/2207.02696.pdf).

Args:
  backbone: backbone model, could be any model with pyramid stage structure.
      Default None for YOLOV7Backbone.
"""

__tail_doc__ = """  features_pick: specific `layer names` or `pyramid feature indexes` from backbone model.
        Default `[-3, -2, -1]` means using the last 3 pyramid feature output from backbone.
  anchors_mode: one of ["efficientdet", "anchor_free", "yolor"], controls which anchor to use.
      - efficientdet anchors default settings: use_object_scores=False, num_anchors=9, anchor_scale=4,
          aspect_ratios=[1, 2, 0.5], num_scales=3, grid_zero_start=False.
      - anchor_free default settings: use_object_scores=True, num_anchors=1, anchor_scale=1,
          aspect_ratios=[1], num_scales=1, grid_zero_start=True.
      - yolor default settings: use_object_scores=True, num_anchors=3.
      Default "yolor".
  num_anchors: number of anchors for a single grid point, should be same with dataset used value.
      Default "auto" means: anchors_mode=="anchor_free" -> 1, anchors_mode=="yolor" -> 3, else 9.
  use_object_scores: bollean value if model header output includes `object_scores`.
      Default "auto" means: True if anchors_mode=="anchor_free" or anchors_mode=="yolor", else False.
  input_shape: input shape if backbone is None, else will use input_shape from backbone.
  num_classes: total output classes. Set `0` to disable `classifier` output. Default 80 for COCO.
  activation: activation used in whole model, default `swish`. Default "swish".
  classifier_activation: The activation function to use for classifier output if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer. Default `sigmoid`.
  freeze_backbone: set `True` for `backbone.trainable = False`. Default `False`.
  pretrained: one of `None` (random initialization) or 'coco' (pre-training on COCO).
      Will try to download and load pre-trained model weights if not None. Default `coco`.
  pyramid_levels_min: anchors inititial parameter for model prediction, not affecting model architecture. Default `3`.
      pyramid_levels_max is calculated as `pyramid_levels_min + len(features_pick) - 1`.
  anchor_scale: anchors inititial parameter for model prediction, not affecting model architecture.
      Default "auto" means: 1 if (anchors_mode=="anchor_free" or anchors_mode=="yolor"), else 4.
  rescale_mode: model precessing input, not for model structure. Defulat "raw01" means input value in range `[0, 1]`.

Returns:
    A `keras.Model` instance.
"""

YOLOV7.__doc__ = __head_doc__ + """
[YOLOV7Backbone parameters]
  csp_channels: YOLOV7Backbone backbone channel for each block, default: `[64, 128, 256, 256]`.
  stack_concats: .
  stack_depth: .
  stack_out_ratio: .
  use_additional_stack: .
  stem_width: CSPDarknet backbone stem width, default -1 means using csp_channels[0] // 2.
  stem_type: One of ["conv3", "focus", "conv1"], "focus" for YOLOV7_*6 models, "conv1" for YOLOV7_Tiny. Default "conv3".
  csp_downsample_ratios: list value controls down sample block channel ratios for each stack.
  spp_depth: .

[FPN parameters]
  fpn_hidden_channels: .
  fpn_channel_ratio: .
  fpn_stack_concats: .
  fpn_stack_depth: .
  fpn_mid_ratio: .
  fpn_csp_downsample_ratio: .
  use_reparam_conv_head: .
  use_csp_downsample: boolean value if CSPDarknet backbone and FPN downsample using `csp_conv_downsample`
      or a conv layer with `kernel_size=3, strides=2`. Default False.
  model_name: string, model name.

[Detector parameters]
""" + __tail_doc__ + """
Model architectures:
  | Model       | Params | FLOPs  | Input | COCO val AP | test AP |
  | ----------- | ------ | ------ | ----- | ----------- | ------- |
  | YOLOV7_Tiny | 6.23M  | 2.90G  | 416   | 33.3        |         |
  | YOLOV7_CSP  | 37.67M | 53.0G  | 640   | 51.4        |         |
  | YOLOV7_X    | 71.41M | 95.0G  | 640   | 53.1        |         |
  | YOLOV7_W6   | 70.49M | 180.1G | 1280  | 54.9        |         |
  | YOLOV7_E6   | 97.33M | 257.6G | 1280  | 56.0        |         |
  | YOLOV7_D6   | 133.9M | 351.4G | 1280  | 56.6        |         |
  | YOLOV7_E6E  | 151.9M | 421.7G | 1280  | 56.8        |         |
"""

YOLOV7_Tiny.__doc__ = __head_doc__ + __tail_doc__
YOLOV7_CSP.__doc__ = YOLOV7_Tiny.__doc__
YOLOV7_X.__doc__ = YOLOV7_Tiny.__doc__
YOLOV7_W6.__doc__ = YOLOV7_Tiny.__doc__
YOLOV7_E6.__doc__ = YOLOV7_Tiny.__doc__
YOLOV7_D6.__doc__ = YOLOV7_Tiny.__doc__
YOLOV7_E6E.__doc__ = YOLOV7_Tiny.__doc__



from keras_cv_attention_models.coatnet.coatnet import (
    CoAtNet,
    CoAtNetT,
    CoAtNet0,
    CoAtNet1,
    CoAtNet2,
    CoAtNet3,
    CoAtNet4,
    CoAtNet5,
    CoAtNet6,
    CoAtNet7,
    mhsa_with_multi_head_relative_position_embedding,
)

__head_doc__ = """
Keras implementation of `CoAtNet`, or `Conv + Transformer` networks.
Paper [PDF 2106.04803 CoAtNet: Marrying Convolution and Attention for All Data Sizes](https://arxiv.org/pdf/2106.04803.pdf).
"""

__tail_doc__ = """  block_types: block types for each stack,
      - `conv` or any `c` / `C` starts word, means `res_MBConv` block.
      - `transfrom` or any not `c` / `C` starts word, means `res_mhsa` + `res_ffn` block.
      value could be in format like `"cctt"` or `"CCTT"` or `["conv", "conv", "transfrom", "transform"]`.
  expansion: filter expansion in each block. The larger the wider.
  se_ratio: value in `(0, 1)`, where `0` means not using `se_module`. Should be a `number` or `list`, indicates `se_ratio` for each stack.
      Each element can also be a `number` or `list`, indicates `se_ratio` for each block.
  head_dimension: heads number for transformer block. `128` for `CoAtNet6` and `CoAtNet7`, `64` for `CoAtNet5`, `32` for others.
  use_dw_strides: boolean value if using strides on `Conv2D` or next `DepthWiseConv2D` for `res_MBConv` block.
      It's claimed higher accuracy for small models using strides on `DepthWiseConv2D`. Default True.
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `relu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  dropout: dropout rate if top layers is included.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
  pretrained: None or "imagenet". Currently only `CoAtNet0` with "imagenet" available.
  **kwargs: other parameters if available.
Returns:
    A `keras.Model` instance.
"""

CoAtNet.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  out_channels: output channels for each stack.
  stem_width: output dimension for stem block.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model                               | Params | FLOPs  | Input | Top1 Acc |
  | ----------------------------------- | ------ | ------ | ----- | -------- |
  | CoAtNet0 (Self trained 105 epochs)  | 23.8M  | 2.17G  | 160   | 80.48    |
  | CoAtNet0 (Self trained 305 epochs)  | 23.8M  | 4.22G  | 224   | 82.79    |
  | CoAtNet0                            | 25M    | 4.2G   | 224   | 81.6     |
  | CoAtNet0, Stride-2 DConv2D          | 25M    | 4.6G   | 224   | 82.0     |
  | CoAtNet0                            | 25M    | 13.4G  | 384   | 83.9     |
  | CoAtNet1                            | 42M    | 8.4G   | 224   | 83.3     |
  | CoAtNet1, Stride-2 DConv2D          | 42M    | 8.8G   | 224   | 83.5     |
  | CoAtNet1                            | 42M    | 27.4G  | 384   | 85.1     |
  | CoAtNet2                            | 75M    | 15.7G  | 224   | 84.1     |
  | CoAtNet2, Stride-2 DConv2D          | 75M    | 16.6G  | 224   | 84.1     |
  | CoAtNet2                            | 75M    | 49.8G  | 384   | 85.7     |
  | CoAtNet2                            | 75M    | 96.7G  | 512   | 85.9     |
  | CoAtNet2, ImageNet-21k pretrain     | 75M    | 16.6G  | 224   | 87.1     |
  | CoAtNet2, ImageNet-21k pretrain     | 75M    | 49.8G  | 384   | 87.1     |
  | CoAtNet2, ImageNet-21k pretrain     | 75M    | 96.7G  | 512   | 87.3     |
  | CoAtNet3                            | 168M   | 34.7G  | 224   | 84.5     |
  | CoAtNet3                            | 168M   | 107.4G | 384   | 85.8     |
  | CoAtNet3                            | 168M   | 203.1G | 512   | 86.0     |
  | CoAtNet3, ImageNet-21k pretrain     | 168M   | 34.7G  | 224   | 87.6     |
  | CoAtNet3, ImageNet-21k pretrain     | 168M   | 107.4G | 384   | 87.6     |
  | CoAtNet3, ImageNet-21k pretrain     | 168M   | 203.1G | 512   | 87.9     |
  | CoAtNet4, ImageNet-21k pretrain     | 275M   | 189.5G | 384   | 87.9     |
  | CoAtNet4, ImageNet-21k pretrain     | 275M   | 360.9G | 512   | 88.1     |
  | CoAtNet4, ImageNet-21K + PT-RA-E150 | 275M   | 189.5G | 384   | 88.4     |
  | CoAtNet4, ImageNet-21K + PT-RA-E150 | 275M   | 360.9G | 512   | 88.56    |

  **JFT pre-trained models accuracy**

  | Model                      | Input | Reported Params    | self-defined Params    | Top1 Acc |
  | -------------------------- | ----- | ------------------ | ---------------------- | -------- |
  | CoAtNet3, Stride-2 DConv2D | 384   | 168M, FLOPs 114G   | 160.64M, FLOPs 109.67G | 88.52    |
  | CoAtNet3, Stride-2 DConv2D | 512   | 168M, FLOPs 214G   | 161.24M, FLOPs 205.06G | 88.81    |
  | CoAtNet4                   | 512   | 275M, FLOPs 361G   | 270.69M, FLOPs 359.77G | 89.11    |
  | CoAtNet5                   | 512   | 688M, FLOPs 812G   | 676.23M, FLOPs 807.06G | 89.77    |
  | CoAtNet6                   | 512   | 1.47B, FLOPs 1521G | 1.336B, FLOPs 1470.56G | 90.45    |
  | CoAtNet7                   | 512   | 2.44B, FLOPs 2586G | 2.413B, FLOPs 2537.56G | 90.88    |
"""

CoAtNet0.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

CoAtNet1.__doc__ = CoAtNet0.__doc__
CoAtNet2.__doc__ = CoAtNet0.__doc__
CoAtNet3.__doc__ = CoAtNet0.__doc__
CoAtNet4.__doc__ = CoAtNet0.__doc__
CoAtNet5.__doc__ = CoAtNet0.__doc__
CoAtNet6.__doc__ = CoAtNet0.__doc__
CoAtNet7.__doc__ = CoAtNet0.__doc__

mhsa_with_multi_head_relative_position_embedding.__doc__ = __head_doc__ + """
Multi head self attention with multi head relative positional embedding. Defined as function, not layer.
Typical MHSA with `MultiHeadRelativePositionalEmbedding` added to `attention_scores`.

Args:
  inputs: input tensor.
  num_heads: Number of attention heads.
  key_dim: Size of each attention head for query and key. Default `0` for `key_dim = inputs.shape[-1] // num_heads`.
  global_query: whether use a specified `query` from outside. Default `None` for not using.
  out_shape: The expected shape of an output tensor. If not specified, projects back to the input dim.
  out_weight: Boolean, whether use an ouput dense.
  qkv_bias: Boolean, whether the qkv dense layer use bias vectors/matrices.
  out_bias: Boolean, whether the ouput dense layer use bias vectors/matrices.
  attn_dropout: Dropout probability for attention.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([14, 16, 256])
>>> nn = attention_layers.mhsa_with_multi_head_relative_position_embedding(inputs, num_heads=4, out_shape=512)
>>> print(f"{nn.shape = }")
# nn.shape = TensorShape([None, 14, 16, 512])

>>> mm = keras.models.Model(inputs, nn)
>>> mm.summary()
>>> print({ii.name: ii.shape for ii in mm.weights})
# {'conv2d_2/kernel:0': TensorShape([1, 1, 256, 1024]),
#  'multi_head_relative_positional_embedding_1/pos_emb:0': TensorShape([837, 4]),
#  'dense_2/kernel:0': TensorShape([512, 512])}
"""
